from igraph._igraph import (
    IN,
    OUT,
    arpack_options as default_arpack_options,
)
from igraph.statistics import Histogram
from igraph.utils import deprecated


def _indegree(graph, *args, **kwds):
    """Returns the in-degrees in a list.

    See L{GraphBase.degree} for possible arguments.
    """
    kwds["mode"] = IN
    return graph.degree(*args, **kwds)


def _outdegree(graph, *args, **kwds):
    """Returns the out-degrees in a list.

    See L{GraphBase.degree} for possible arguments.
    """
    kwds["mode"] = OUT
    return graph.degree(*args, **kwds)


def _degree_distribution(graph, bin_width=1, *args, **kwds):
    """Calculates the degree distribution of the graph.

    Unknown keyword arguments are directly passed to L{GraphBase.degree}.

    @param bin_width: the bin width of the histogram
    @return: a histogram representing the degree distribution of the
      graph.
    """
    result = Histogram(bin_width, graph.degree(*args, **kwds))
    return result


def _pagerank(
    graph,
    vertices=None,
    directed=True,
    damping=0.85,
    weights=None,
    arpack_options=None,
    implementation="prpack",
):
    """Calculates the PageRank values of a graph.

    @param vertices: the indices of the vertices being queried.
      C{None} means all of the vertices.
    @param directed: whether to consider directed paths.
    @param damping: the damping factor. M{1-damping} is the probability of
      resetting the random walk to a uniform distribution in each step.
    @param weights: edge weights to be used. Can be a sequence or iterable
      or even an edge attribute name.
    @param arpack_options: an L{ARPACKOptions} object used to fine-tune
      the ARPACK eigenvector calculation. If omitted, the module-level
      variable called C{arpack_options} is used. This argument is
      ignored if not the ARPACK implementation is used, see the
      I{implementation} argument.
    @param implementation: which implementation to use to solve the
      PageRank eigenproblem. Possible values are:
        - C{"prpack"}: use the PRPACK library. This is a new
          implementation in igraph 0.7
        - C{"arpack"}: use the ARPACK library. This implementation
          was used from version 0.5, until version 0.7.
    @return: a list with the PageRank values of the specified vertices.
    """
    if arpack_options is None:
        arpack_options = default_arpack_options
    return graph.personalized_pagerank(
        vertices,
        directed,
        damping,
        None,
        None,
        weights,
        arpack_options,
        implementation,
    )


def _shortest_paths(graph, *args, **kwds):
    """Deprecated alias to L{Graph.distances()}."""
    deprecated("Graph.shortest_paths() is deprecated; use Graph.distances() instead")
    return graph.distances(*args, **kwds)



"""This module provides implementation for a Cairo-specific matrix drawer."""

from itertools import islice
from math import pi

from igraph.drawing.cairo.base import AbstractCairoDrawer

__all__ = ("CairoMatrixDrawer",)


class CairoMatrixDrawer(AbstractCairoDrawer):
    """Default Cairo drawer object for matrices."""

    def __init__(self, context):
        """Constructs the vertex drawer and associates it to the given
        palette.

        @param context: the context on which we will draw
        """
        super().__init__(context, bbox=None)

    def draw(self, matrix, **kwds):
        """Draws the given Matrix in a Cairo context.

        @param matrix: the igraph.Matrix to plot.

        It accepts the following keyword arguments:

          - C{bbox}:    the bounding box within which we will draw.
            Can be anything accepted by the constructor of L{BoundingBox}
            (i.e., a 2-tuple, a 4-tuple or a L{BoundingBox} object).

          - C{palette}: the palette that can be used to map integer color
            indices to colors when drawing vertices

          - C{style}: the style of the plot. C{boolean} is useful for plotting
            matrices with boolean (C{True}/C{False} or 0/1) values: C{False}
            will be shown with a white box and C{True} with a black box.
            C{palette} uses the given palette to represent numbers by colors,
            the minimum will be assigned to palette color index 0 and the maximum
            will be assigned to the length of the palette. C{None} draws transparent
            cell backgrounds only. The default style is C{boolean} (but it may
            change in the future). C{None} values in the matrix are treated
            specially in both cases: nothing is drawn in the cell corresponding
            to C{None}.

          - C{square}: whether the cells of the matrix should be square or not.
            Default is C{True}.

          - C{grid_width}: line width of the grid shown on the matrix. If zero or
            negative, the grid is turned off. The grid is also turned off if the size
            of a cell is less than three times the given line width. Default is C{1}.
            Fractional widths are also allowed.

          - C{border_width}: line width of the border drawn around the matrix.
            If zero or negative, the border is turned off. Default is C{1}.

          - C{row_names}: the names of the rows

          - C{col_names}: the names of the columns.

          - C{values}: values to be displayed in the cells. If C{None} or
            C{False}, no values are displayed. If C{True}, the values come
            from the matrix being plotted. If it is another matrix, the
            values of that matrix are shown in the cells. In this case,
            the shape of the value matrix must match the shape of the
            matrix being plotted.

          - C{value_format}: a format string or a callable that specifies how
            the values should be plotted. If it is a callable, it must be a
            function that expects a single value and returns a string.
            Example: C{"%#.2f"} for floating-point numbers with always exactly
            two digits after the decimal point. See the Python documentation of
            the C{%} operator for details on the format string. If the format
            string is not given, it defaults to the C{str} function.

        If only the row names or the column names are given and the matrix
        is square-shaped, the same names are used for both column and row
        names.
        """
        context = self.context
        Matrix = matrix.__class__

        bbox = self.bbox = kwds.pop("bbox", None)
        palette = kwds.pop("palette", None)
        if bbox is None:
            raise ValueError("bbox is required for Cairo plots")
        if palette is None:
            raise ValueError("palette is required for Cairo plots")

        grid_width = float(kwds.get("grid_width", 1.0))
        border_width = float(kwds.get("border_width", 1.0))
        style = kwds.get("style", "boolean")
        row_names = kwds.get("row_names")
        col_names = kwds.get("col_names", row_names)
        values = kwds.get("values")
        value_format = kwds.get("value_format", str)

        # Validations
        if style not in ("boolean", "palette", "none", None):
            raise ValueError("invalid style")
        if style == "none":
            style = None
        if row_names is None and col_names is not None:
            row_names = col_names
        if row_names is not None:
            row_names = [str(name) for name in islice(row_names, matrix._nrow)]
            if len(row_names) < matrix._nrow:
                row_names.extend([""] * (matrix._nrow - len(row_names)))
        if col_names is not None:
            col_names = [str(name) for name in islice(col_names, matrix._ncol)]
            if len(col_names) < matrix._ncol:
                col_names.extend([""] * (matrix._ncol - len(col_names)))
        if values is False:
            values = None
        if values is True:
            values = matrix
        if isinstance(values, list):
            values = Matrix(list)
        if values is not None and not isinstance(values, Matrix):
            raise TypeError("values must be None, False, True or a matrix")
        if values is not None and values.shape != matrix.shape:
            raise ValueError("values must be a matrix of size %s" % matrix.shape)

        # Calculate text extents if needed
        if row_names is not None or col_names is not None:
            te = context.text_extents
            space_width = te(" ")[4]
            if row_names is not None:
                max_row_name_width = max([te(s)[4] for s in row_names]) + space_width
            else:
                max_row_name_width = 0
            if col_names is not None:
                max_col_name_width = max([te(s)[4] for s in col_names]) + space_width
            else:
                max_col_name_width = 0
        else:
            max_row_name_width, max_col_name_width = 0, 0
            space_width = 0

        # Calculate sizes
        total_width = float(bbox.width) - max_row_name_width
        total_height = float(bbox.height) - max_col_name_width
        dx = total_width / matrix.shape[1]
        dy = total_height / matrix.shape[0]
        if kwds.get("square", True):
            dx, dy = min(dx, dy), min(dx, dy)
        total_width, total_height = dx * matrix.shape[1], dy * matrix.shape[0]
        ox = bbox.left + (bbox.width - total_width - max_row_name_width) / 2.0
        oy = bbox.top + (bbox.height - total_height - max_col_name_width) / 2.0
        ox += max_row_name_width
        oy += max_col_name_width

        # Determine rescaling factors for the palette if needed
        if style == "palette":
            mi, ma = matrix.min(), matrix.max()
            color_offset = mi
            color_ratio = (len(palette) - 1) / float(ma - mi)
        else:
            color_offset, color_ratio = 0, 1

        # Validate grid width
        if dx < 3 * grid_width or dy < 3 * grid_width:
            grid_width = 0.0
        if grid_width > 0:
            context.set_line_width(grid_width)
        else:
            # When the grid width is zero, we will still stroke the
            # rectangles, but with the same color as the fill color
            # of the cell - otherwise we would get thin white lines
            # between the cells as a drawing artifact
            context.set_line_width(1)

        # Draw row names (if any)
        context.set_source_rgb(0.0, 0.0, 0.0)
        if row_names is not None:
            x, y = ox, oy
            for heading in row_names:
                _, _, _, h, xa, _ = context.text_extents(heading)
                context.move_to(x - xa - space_width, y + (dy + h) / 2.0)
                context.show_text(heading)
                y += dy

        # Draw column names (if any)
        if col_names is not None:
            context.save()
            context.translate(ox, oy)
            context.rotate(-pi / 2)
            x, y = 0.0, 0.0
            for heading in col_names:
                _, _, _, h, _, _ = context.text_extents(heading)
                context.move_to(x + space_width, y + (dx + h) / 2.0)
                context.show_text(heading)
                y += dx
            context.restore()

        # Draw matrix
        x, y = ox, oy
        if style is None:
            fill = lambda: None  # noqa: E731
        else:
            fill = context.fill_preserve
        for row in matrix:
            for item in row:
                if item is None:
                    x += dx
                    continue
                if style == "boolean":
                    if item:
                        context.set_source_rgb(0.0, 0.0, 0.0)
                    else:
                        context.set_source_rgb(1.0, 1.0, 1.0)
                elif style == "palette":
                    cidx = int((item - color_offset) * color_ratio)
                    if cidx < 0:
                        cidx = 0
                    context.set_source_rgba(*palette.get(cidx))
                context.rectangle(x, y, dx, dy)
                if grid_width > 0:
                    fill()
                    context.set_source_rgb(0.5, 0.5, 0.5)
                    context.stroke()
                else:
                    fill()
                    context.stroke()
                x += dx
            x, y = ox, y + dy

        # Draw cell values
        if values is not None:
            x, y = ox, oy
            context.set_source_rgb(0.0, 0.0, 0.0)
            for row in values.data:
                if callable(value_format):
                    values = [value_format(item) for item in row]
                else:
                    values = [value_format % item for item in row]
                for item in values:
                    th, tw = context.text_extents(item)[3:5]
                    context.move_to(x + (dx - tw) / 2.0, y + (dy + th) / 2.0)
                    context.show_text(item)
                    x += dx
                x, y = ox, y + dy

        # Draw borders
        if border_width > 0:
            context.set_line_width(border_width)
            context.set_source_rgb(0.0, 0.0, 0.0)
            context.rectangle(ox, oy, dx * matrix.shape[1], dy * matrix.shape[0])
            context.stroke()



"""
Drawing and plotting routines for IGraph.

igraph has two plotting backends at the moment: Cairo and Matplotlib.

The Cairo backend is dependent on the C{pycairo} or C{cairocffi} libraries that
provide Python bindings to the popular U{Cairo library<http://www.cairographics.org>}.
This means that if you don't have U{pycairo<http://www.cairographics.org/pycairo>}
or U{cairocffi<https://doc.courtbouillon.org/cairocffi/>} installed, you won't be able
to use the Cairo plotting backend. Whenever the documentation refers to the
C{pycairo} library, you can safely replace it with C{cairocffi} as the two are
API-compatible.

The Matplotlib backend uses the U{Matplotlib library<https://matplotlib.org>}.
You will need to install it from PyPI if you want to use the Matplotlib
plotting backend.

If you do not want to (or cannot) install any of the dependencies outlined
above, you can still save the graph to an SVG file and view it from
U{Mozilla Firefox<https://www.mozilla.org/firefox>} (free) or edit it in
U{Inkscape<https://www.inkscape.org>} (free), U{Skencil<https://www.skencil.org>}
(formerly known as Sketch, also free) or Adobe Illustrator.
"""


import os

from io import BytesIO
from warnings import warn

from igraph.configuration import Configuration
from igraph.drawing.cairo.utils import find_cairo
from igraph.drawing.colors import Palette, palettes
from igraph.drawing.utils import BoundingBox
from igraph.utils import named_temporary_file

__all__ = ("CairoPlot",)

cairo = find_cairo()

#####################################################################


class CairoPlot:
    """Class representing an arbitrary plot that uses the Cairo plotting
    backend.

    Objects that you can plot include graphs, matrices, palettes, clusterings,
    covers, and dendrograms.

    In Cairo, every plot has an associated surface object. The surface is an
    instance of C{cairo.Surface}, a member of the C{pycairo} library. The
    surface itself provides a unified API to various plotting targets like SVG
    files, X11 windows, PostScript files, PNG files and so on. C{igraph} does
    not usually know on which surface it is plotting at each point in time,
    since C{pycairo} takes care of the actual drawing. Everything that's
    supported by C{pycairo} should be supported by this class as well.

    Current Cairo surfaces include:

      - C{cairo.GlitzSurface} -- OpenGL accelerated surface for the X11
        Window System.

      - C{cairo.ImageSurface} -- memory buffer surface. Can be written to a
        C{PNG} image file.

      - C{cairo.PDFSurface} -- PDF document surface.

      - C{cairo.PSSurface} -- PostScript document surface.

      - C{cairo.SVGSurface} -- SVG (Scalable Vector Graphics) document surface.

      - C{cairo.Win32Surface} -- Microsoft Windows screen rendering.

      - C{cairo.XlibSurface} -- X11 Window System screen rendering.

    If you create a C{Plot} object with a string given as the target surface,
    the string will be treated as a filename, and its extension will decide
    which surface class will be used. Please note that not all surfaces might
    be available, depending on your C{pycairo} installation.

    A C{Plot} has an assigned default palette (see L{igraph.drawing.colors.Palette})
    which is used for plotting objects.

    A C{Plot} object also has a list of objects to be plotted with their
    respective bounding boxes, palettes and opacities. Palettes assigned to an
    object override the default palette of the plot. Objects can be added by the
    L{Plot.add} method and removed by the L{Plot.remove} method.
    """

    def __init__(
        self,
        target=None,
        bbox=None,
        palette=None,
        background=None,
    ):
        """Creates a new plot.

        @param target: the target surface to write to. It can be one of the
          following types:

            - C{None} -- a Cairo surface will be created and the object will be
              plotted there.

            - C{cairo.Surface} -- the given Cairo surface will be used.

            - C{string} -- a file with the given name will be created and an
              appropriate Cairo surface will be attached to it.

        @param bbox: the bounding box of the surface. It is interpreted
          differently with different surfaces: PDF and PS surfaces will treat it
          as points (1 point = 1/72 inch). Image surfaces will treat it as
          pixels. SVG surfaces will treat it as an abstract unit, but it will
          mostly be interpreted as pixels when viewing the SVG file in Firefox.

        @param palette: the palette primarily used on the plot if the
          added objects do not specify a private palette. Must be either
          an L{igraph.drawing.colors.Palette} object or a string referring
          to a valid key of C{igraph.drawing.colors.palettes} (see module
          L{igraph.drawing.colors}) or C{None}. In the latter case, the default
          palette given by the configuration key C{plotting.palette} is used.

        @param background: the background color. If C{None}, the background
          will be transparent. You can use any color specification here that is
          understood by L{igraph.drawing.colors.color_name_to_rgba}.
        """

        self._filename = None
        self._need_tmpfile = False

        if bbox is None:
            self.bbox = BoundingBox(600, 600)
        elif isinstance(bbox, tuple) or isinstance(bbox, list):
            self.bbox = BoundingBox(bbox)
        else:
            self.bbox = bbox

        if palette is None:
            config = Configuration.instance()
            palette = config["plotting.palette"]
        if not isinstance(palette, Palette):
            palette = palettes[palette]
        self._palette = palette

        if target is None:
            self._need_tmpfile = True
            self._surface = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, int(self.bbox.width), int(self.bbox.height)
            )
        elif isinstance(target, cairo.Surface):
            self._surface = target
        else:
            self._filename = target
            _, ext = os.path.splitext(target)
            ext = ext.lower()
            if ext == ".pdf":
                self._surface = cairo.PDFSurface(
                    target, self.bbox.width, self.bbox.height
                )
            elif ext == ".ps" or ext == ".eps":
                self._surface = cairo.PSSurface(
                    target, self.bbox.width, self.bbox.height
                )
            elif ext == ".png":
                self._surface = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, int(self.bbox.width), int(self.bbox.height)
                )
            elif ext == ".svg":
                self._surface = cairo.SVGSurface(
                    target, self.bbox.width, self.bbox.height
                )
            else:
                raise ValueError("image format not handled by Cairo: %s" % ext)

        self._ctx = cairo.Context(self._surface)

        self._objects = []
        self._is_dirty = False

        if background is None:
            background = "white"
        self.background = background

    def add(self, obj, bbox=None, palette=None, opacity=1.0, *args, **kwds):
        """Adds an object to the plot.

        Arguments not specified here are stored and passed to the object's
        plotting function when necessary. Since you are most likely interested
        in the arguments acceptable by graphs, see L{Graph.__plot__} for more
        details.

        @param obj: the object to be added
        @param bbox: the bounding box of the object. If C{None}, the object
          will fill the entire area of the plot.
        @param palette: the color palette used for drawing the object. If the
          object tries to get a color assigned to a positive integer, it
          will use this palette. If C{None}, defaults to the global palette
          of the plot.
        @param opacity: the opacity of the object being plotted, in the range
          0.0-1.0

        @see: Graph.__plot__
        """
        if opacity < 0.0 or opacity > 1.0:
            raise ValueError("opacity must be between 0.0 and 1.0")
        if bbox is None:
            bbox = self.bbox
        if (bbox is not None) and (not isinstance(bbox, BoundingBox)):
            bbox = BoundingBox(bbox)
        self._objects.append((obj, bbox, palette, opacity, args, kwds))
        self.mark_dirty()

    @property
    def background(self):
        """Returns the background color of the plot. C{None} means a
        transparent background.
        """
        return self._background

    @background.setter
    def background(self, color):
        """Sets the background color of the plot. C{None} means a
        transparent background. You can use any color specification here
        that is understood by the C{get} method of the current palette
        or by L{igraph.drawing.colors.color_name_to_rgb}.
        """
        if color is None:
            self._background = None
        else:
            self._background = self._palette.get(color)

    def remove(self, obj, bbox=None, idx=1):
        """Removes an object from the plot.

        If the object has been added multiple times and no bounding box
        was specified, it removes the instance which occurs M{idx}th
        in the list of identical instances of the object.

        @param obj: the object to be removed
        @param bbox: optional bounding box specification for the object.
          If given, only objects with exactly this bounding box will be
          considered.
        @param idx: if multiple objects match the specification given by
          M{obj} and M{bbox}, only the M{idx}th occurrence will be removed.
        @return: C{True} if the object has been removed successfully,
          C{False} if the object was not on the plot at all or M{idx}
          was larger than the count of occurrences
        """
        for i in range(len(self._objects)):
            current_obj, current_bbox = self._objects[i][0:2]
            if current_obj is obj and (bbox is None or current_bbox == bbox):
                idx -= 1
                if idx == 0:
                    self._objects[i : (i + 1)] = []
                    self.mark_dirty()
                    return True
        return False

    def mark_dirty(self):
        """Marks the plot as dirty (should be redrawn)"""
        self._is_dirty = True

    def redraw(self, context=None):
        """Redraws the plot"""
        ctx = context or self._ctx
        if self._background is not None:
            ctx.set_source_rgba(*self._background)
            ctx.rectangle(0, 0, self.bbox.width, self.bbox.height)
            ctx.fill()

        for obj, bbox, palette, opacity, args, kwds in self._objects:
            if palette is None:
                palette = getattr(obj, "_default_palette", self._palette)
            plotter = getattr(obj, "__plot__", None)
            if plotter is None:
                warn("%s does not support plotting" % (obj,), stacklevel=1)
            else:
                if opacity < 1.0:
                    ctx.push_group()
                else:
                    ctx.save()
                plotter(
                    "cairo",
                    ctx,
                    bbox=bbox,
                    palette=palette,
                    *args,  # noqa: B026
                    **kwds
                )
                if opacity < 1.0:
                    ctx.pop_group_to_source()
                    ctx.paint_with_alpha(opacity)
                else:
                    ctx.restore()

        self._is_dirty = False

    def save(self, fname=None):
        """Saves the plot.

        @param fname: the filename to save to. It is ignored if the surface
          of the plot is not an C{ImageSurface}.
        """
        if self._is_dirty:
            self.redraw()
        if isinstance(self._surface, cairo.ImageSurface):
            if fname is None and self._need_tmpfile:
                with named_temporary_file(prefix="igraph", suffix=".png") as fname:
                    self._surface.write_to_png(fname)
                    return None

            fname = fname or self._filename
            if fname is None:
                raise ValueError("no file name is known for the surface and none given")

            # Conversion to string is needed because the user might pass a Path
            # object and cairocffi expects a string
            return self._surface.write_to_png(str(fname))

        if fname is not None:
            warn(
                "filename is ignored for surfaces other than ImageSurface", stacklevel=1
            )

        self._ctx.show_page()
        self._surface.finish()

    def _repr_svg_(self):
        """Returns an SVG representation of this plot as a string.

        This method is used by IPython to display this plot inline.
        """
        io = BytesIO()
        # Create a new SVG surface and use that to get the SVG representation,
        # which will end up in io
        surface = cairo.SVGSurface(io, self.bbox.width, self.bbox.height)
        context = cairo.Context(surface)
        # Plot the graph on this context
        self.redraw(context)
        # No idea why this is needed but python crashes without
        context.show_page()
        surface.finish()
        # Return the raw SVG representation
        result = io.getvalue().decode("utf-8")
        return result, {"isolated": True}  # put it inside an iframe

    @property
    def bounding_box(self):
        """Returns the bounding box of the Cairo surface as a
        L{BoundingBox} object"""
        return BoundingBox(self.bbox)

    @property
    def height(self):
        """Returns the height of the Cairo surface on which the plot
        is drawn"""
        return self.bbox.height

    @property
    def surface(self):
        """Returns the Cairo surface on which the plot is drawn"""
        return self._surface

    @property
    def width(self):
        """Returns the width of the Cairo surface on which the plot
        is drawn"""
        return self.bbox.width



import unittest

from igraph import Graph, InternalError


try:
    import numpy as np
except ImportError:
    np = None

try:
    import scipy.sparse as sparse
except ImportError:
    sparse = None

try:
    import pandas as pd
except ImportError:
    pd = None


class GeneratorTests(unittest.TestCase):
    def testStar(self):
        g = Graph.Star(5, "in")
        el = [(1, 0), (2, 0), (3, 0), (4, 0)]
        self.assertTrue(g.is_directed())
        self.assertTrue(g.get_edgelist() == el)
        g = Graph.Star(5, "out", center=2)
        el = [(2, 0), (2, 1), (2, 3), (2, 4)]
        self.assertTrue(g.is_directed())
        self.assertTrue(g.get_edgelist() == el)
        g = Graph.Star(5, "mutual", center=2)
        el = [(0, 2), (1, 2), (2, 0), (2, 1), (2, 3), (2, 4), (3, 2), (4, 2)]
        self.assertTrue(g.is_directed())
        self.assertTrue(sorted(g.get_edgelist()) == el)
        g = Graph.Star(5, center=3)
        el = [(0, 3), (1, 3), (2, 3), (3, 4)]
        self.assertTrue(not g.is_directed())
        self.assertTrue(sorted(g.get_edgelist()) == el)

    def testFamous(self):
        g = Graph.Famous("tutte")
        self.assertTrue(g.vcount() == 46 and g.ecount() == 69)
        self.assertRaises(InternalError, Graph.Famous, "unknown")

    def testFormula(self):
        tests = [
            (None, [], []),
            ("", [""], []),
            ("A", ["A"], []),
            ("A-B", ["A", "B"], [(0, 1)]),
            ("A --- B", ["A", "B"], [(0, 1)]),
            (
                "A--B, C--D, E--F, G--H, I, J, K",
                ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K"],
                [(0, 1), (2, 3), (4, 5), (6, 7)],
            ),
            (
                "A:B:C:D -- A:B:C:D",
                ["A", "B", "C", "D"],
                [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)],
            ),
            ("A -> B -> C", ["A", "B", "C"], [(0, 1), (1, 2)]),
            ("A <- B -> C", ["A", "B", "C"], [(1, 0), (1, 2)]),
            ("A <- B -- C", ["A", "B", "C"], [(1, 0)]),
            (
                "A <-> B <---> C <> D",
                ["A", "B", "C", "D"],
                [(0, 1), (1, 0), (1, 2), (2, 1), (2, 3), (3, 2)],
            ),
            (
                "'this is' <- 'a silly' -> 'graph here'",
                ["this is", "a silly", "graph here"],
                [(1, 0), (1, 2)],
            ),
            (
                "Alice-Bob-Cecil-Alice, Daniel-Cecil-Eugene, Cecil-Gordon",
                ["Alice", "Bob", "Cecil", "Daniel", "Eugene", "Gordon"],
                [(0, 1), (1, 2), (0, 2), (2, 3), (2, 4), (2, 5)],
            ),
            (
                "Alice-Bob:Cecil:Daniel, Cecil:Daniel-Eugene:Gordon",
                ["Alice", "Bob", "Cecil", "Daniel", "Eugene", "Gordon"],
                [(0, 1), (0, 2), (0, 3), (2, 4), (2, 5), (3, 4), (3, 5)],
            ),
            (
                "Alice <-> Bob --> Cecil <-- Daniel, Eugene --> Gordon:Helen",
                ["Alice", "Bob", "Cecil", "Daniel", "Eugene", "Gordon", "Helen"],
                [(0, 1), (1, 0), (1, 2), (3, 2), (4, 5), (4, 6)],
            ),
            (
                "Alice -- Bob -- Daniel, Cecil:Gordon, Helen",
                ["Alice", "Bob", "Daniel", "Cecil", "Gordon", "Helen"],
                [(0, 1), (1, 2)],
            ),
            (
                '"+" -- "-", "*" -- "/", "%%" -- "%/%"',
                ["+", "-", "*", "/", "%%", "%/%"],
                [(0, 1), (2, 3), (4, 5)],
            ),
            ("A-B-C\nC-D", ["A", "B", "C", "D"], [(0, 1), (1, 2), (2, 3)]),
            ("A-B-C\n    C-D", ["A", "B", "C", "D"], [(0, 1), (1, 2), (2, 3)]),
        ]
        for formula, names, edges in tests:
            g = Graph.Formula(formula)
            self.assertEqual(g.vs["name"], names)
            self.assertEqual(g.get_edgelist(), sorted(edges))

    def testFull(self):
        g = Graph.Full(20, directed=True)
        el = g.get_edgelist()
        el.sort()
        self.assertTrue(g.is_complete())
        self.assertTrue(
            g.get_edgelist() == [(x, y) for x in range(20) for y in range(20) if x != y]
        )

    def testFullCitation(self):
        g = Graph.Full_Citation(20)
        el = g.get_edgelist()
        el.sort()
        self.assertTrue(not g.is_directed())
        self.assertTrue(el == [(x, y) for x in range(19) for y in range(x + 1, 20)])

        g = Graph.Full_Citation(20, True)
        el = g.get_edgelist()
        el.sort()
        self.assertTrue(g.is_directed())
        self.assertTrue(el == [(x, y) for x in range(1, 20) for y in range(x)])

        self.assertRaises(ValueError, Graph.Full_Citation, -2)

    def testHexagonalLattice(self):
        el = [
            (0, 1),
            (0, 6),
            (1, 2),
            (2, 3),
            (2, 8),
            (3, 4),
            (4, 10),
            (5, 6),
            (5, 11),
            (6, 7),
            (7, 8),
            (7, 13),
            (8, 9),
            (9, 10),
            (9, 15),
            (11, 12),
            (12, 13),
            (13, 14),
            (14, 15),
        ]
        g = Graph.Hexagonal_Lattice([2, 2])
        self.assertEqual(sorted(g.get_edgelist()), el)

        g = Graph.Hexagonal_Lattice([2, 2], directed=True, mutual=False)
        self.assertEqual(sorted(g.get_edgelist()), el)

        g = Graph.Hexagonal_Lattice([2, 2], directed=True, mutual=True)
        self.assertEqual(sorted(g.get_edgelist()), sorted(el + [(y, x) for x, y in el]))

    def testHypercube(self):
        el = [(0, 1), (0, 2), (0, 4), (1, 3), (1, 5), (2, 3), (2, 6), (3, 7), (4, 5), (4, 6), (5, 7), (6, 7)]
        g = Graph.Hypercube(3)
        self.assertEqual(g.get_edgelist(), el)

    def testLCF(self):
        g1 = Graph.LCF(12, (5, -5), 6)
        g2 = Graph.Famous("Franklin")
        self.assertTrue(g1.isomorphic(g2))
        self.assertRaises(ValueError, Graph.LCF, 12, (5, -5), -3)

    def testRealizeDegreeSequence(self):
        # Test case insensitivity of options too
        g = Graph.Realize_Degree_Sequence(
            [1, 1],
            None,
            "simPLE",
            "smallest",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == [1, 1])

        # Not implemented, should fail
        self.assertRaises(
            NotImplementedError,
            Graph.Realize_Degree_Sequence,
            [1, 1],
            None,
            "loops",
            "largest",
        )

        g = Graph.Realize_Degree_Sequence(
            [1, 1],
            None,
            "all",
            "largest",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == [1, 1])

        g = Graph.Realize_Degree_Sequence(
            [1, 1],
            None,
            "multi",
            "index",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == [1, 1])

        g = Graph.Realize_Degree_Sequence(
            [1, 1],
            [1, 1],
            "simple",
            "largest",
        )
        self.assertTrue(g.is_directed())
        self.assertTrue(g.indegree() == [1, 1])
        self.assertTrue(g.outdegree() == [1, 1])

        # Not implemented, should fail
        self.assertRaises(
            NotImplementedError,
            Graph.Realize_Degree_Sequence,
            [1, 1],
            [1, 1],
            "multi",
            "largest",
        )

        self.assertRaises(
            ValueError,
            Graph.Realize_Degree_Sequence,
            [1, 1],
            [1, 1],
            "should_fail",
            "index",
        )
        self.assertRaises(
            ValueError,
            Graph.Realize_Degree_Sequence,
            [1, 1],
            [1, 1],
            "multi",
            "should_fail",
        )

        # Degree sequence of Zachary karate club, using optional arguments
        zachary = Graph.Famous("zachary")
        degrees = zachary.degree()
        g = Graph.Realize_Degree_Sequence(degrees)
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == degrees)

    def testRealizeBipartiteDegreeSequence(self):
        deg1 = [2, 2]
        deg2 = [1, 1, 2]
        g = Graph.Realize_Bipartite_Degree_Sequence(
            deg1,
            deg2,
            "simple",
            "smallest",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.is_connected())
        self.assertTrue(g.degree() == deg1 + deg2)

        g = Graph.Realize_Bipartite_Degree_Sequence(
            deg1,
            deg2,
            "simple",
            "largest",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == deg1 + deg2)

        g = Graph.Realize_Bipartite_Degree_Sequence(
            deg1,
            deg2,
            "simple",
            "index",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.degree() == deg1 + deg2)

        deg1 = [3, 1, 1]
        deg2 = [2, 3]
        self.assertRaises(
            InternalError,
            Graph.Realize_Bipartite_Degree_Sequence,
            deg1,
            deg2,
            "simple",
            "smallest",
        )

        self.assertRaises(
            InternalError,
            Graph.Realize_Bipartite_Degree_Sequence,
            deg1,
            deg2,
            "simple",
            "index",
        )

        g = Graph.Realize_Bipartite_Degree_Sequence(
            deg1,
            deg2,
            "multi",
            "smallest",
        )
        self.assertFalse(g.is_directed())
        self.assertTrue(g.is_connected())
        self.assertTrue(g.degree() == deg1 + deg2)

    def testKautz(self):
        g = Graph.Kautz(2, 2)
        deg_in = g.degree(mode="in")
        deg_out = g.degree(mode="out")
        # This is not a proper test, but should spot most errors
        self.assertTrue(g.is_directed() and deg_in == [2] * 12 and deg_out == [2] * 12)

    def testDeBruijn(self):
        g = Graph.De_Bruijn(2, 3)
        deg_in = g.degree(mode="in", loops=True)
        deg_out = g.degree(mode="out", loops=True)
        # This is not a proper test, but should spot most errors
        self.assertTrue(g.is_directed() and deg_in == [2] * 8 and deg_out == [2] * 8)

    def testLattice(self):
        g = Graph.Lattice([4, 3], circular=False)
        self.assertEqual(
            sorted(sorted(x) for x in g.get_edgelist()),
            [
                [0, 1],
                [0, 4],
                [1, 2],
                [1, 5],
                [2, 3],
                [2, 6],
                [3, 7],
                [4, 5],
                [4, 8],
                [5, 6],
                [5, 9],
                [6, 7],
                [6, 10],
                [7, 11],
                [8, 9],
                [9, 10],
                [10, 11],
            ],
        )

        g = Graph.Lattice([4, 3], circular=True)
        self.assertEqual(
            sorted(sorted(x) for x in g.get_edgelist()),
            [
                [0, 1],
                [0, 3],
                [0, 4],
                [0, 8],
                [1, 2],
                [1, 5],
                [1, 9],
                [2, 3],
                [2, 6],
                [2, 10],
                [3, 7],
                [3, 11],
                [4, 5],
                [4, 7],
                [4, 8],
                [5, 6],
                [5, 9],
                [6, 7],
                [6, 10],
                [7, 11],
                [8, 9],
                [8, 11],
                [9, 10],
                [10, 11],
            ],
        )

        g = Graph.Lattice([4, 3], circular=(False, 1))
        self.assertEqual(
            sorted(sorted(x) for x in g.get_edgelist()),
            [
                [0, 1],
                [0, 4],
                [0, 8],
                [1, 2],
                [1, 5],
                [1, 9],
                [2, 3],
                [2, 6],
                [2, 10],
                [3, 7],
                [3, 11],
                [4, 5],
                [4, 8],
                [5, 6],
                [5, 9],
                [6, 7],
                [6, 10],
                [7, 11],
                [8, 9],
                [9, 10],
                [10, 11],
            ],
        )

    def testSBM(self):
        pref_matrix = [[0.5, 0, 0], [0, 0, 0.5], [0, 0.5, 0]]
        n = 60
        types = [20, 20, 20]
        g = Graph.SBM(n, pref_matrix, types)

        # Simple smoke tests for the expected structure of the graph
        self.assertTrue(g.is_simple())
        self.assertFalse(g.is_directed())
        self.assertEqual([0] * 20 + [1] * 40, g.connected_components().membership)
        g2 = g.subgraph(list(range(20, 60)))
        self.assertTrue(not any(e.source // 20 == e.target // 20 for e in g2.es))

        # Check loops argument
        g = Graph.SBM(n, pref_matrix, types, loops=True)
        self.assertFalse(g.is_simple())
        self.assertTrue(sum(g.is_loop()) > 0)

        # Check directedness
        g = Graph.SBM(n, pref_matrix, types, directed=True)
        self.assertTrue(g.is_directed())
        self.assertTrue(sum(g.is_mutual()) < g.ecount())
        self.assertTrue(sum(g.is_loop()) == 0)

        # Check error conditions
        self.assertRaises(ValueError, Graph.SBM, -1, pref_matrix, types)
        self.assertRaises(InternalError, Graph.SBM, 61, pref_matrix, types)
        pref_matrix[0][1] = 0.7
        self.assertRaises(InternalError, Graph.SBM, 60, pref_matrix, types)

    def testTriangularLattice(self):
        g = Graph.Triangular_Lattice([2, 2])
        self.assertEqual(
            sorted(g.get_edgelist()), [(0, 1), (0, 2), (0, 3), (1, 3), (2, 3)]
        )

        g = Graph.Triangular_Lattice([2, 2], directed=True, mutual=False)
        self.assertEqual(
            sorted(g.get_edgelist()), [(0, 1), (0, 2), (0, 3), (1, 3), (2, 3)]
        )

        g = Graph.Triangular_Lattice([2, 2], directed=True, mutual=True)
        self.assertEqual(
            sorted(g.get_edgelist()),
            [
                (0, 1),
                (0, 2),
                (0, 3),
                (1, 0),
                (1, 3),
                (2, 0),
                (2, 3),
                (3, 0),
                (3, 1),
                (3, 2),
            ],
        )

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testAdjacencyNumPy(self):
        mat = np.array(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 2, 0], [0, 1, 0, 0]],
        )

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (2, 2), (3, 1)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (2, 2), (2, 2)])

        # ADJ MAX
        g = Graph.Adjacency(mat, mode="max")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 3), (2, 2), (2, 2)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2), (2, 2), (1, 3)])

        # ADJ UPPER
        g = Graph.Adjacency(mat, mode="upper")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2), (2, 2)])

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testAdjacencyNumPyLoopHandling(self):
        mat = np.array(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 2, 0], [0, 1, 0, 0]],
        )

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (2, 2), (3, 1)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min", loops="twice")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (2, 2)])

        # ADJ MAX
        g = Graph.Adjacency(mat, mode="max", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 3), (2, 2)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2), (1, 3)])

        # ADJ UPPER
        g = Graph.Adjacency(mat, mode="upper", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2)])

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat, loops=False)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (3, 1)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min", loops=False)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1)])

        # ADJ MAX
        g = Graph.Adjacency(mat, mode="max", loops=False)
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 3)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower", loops=False)
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (1, 3)])

        # ADJ UPPER
        g = Graph.Adjacency(mat, mode="upper", loops=False)
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2)])

    @unittest.skipIf(
        (sparse is None) or (np is None), "test case depends on NumPy/SciPy"
    )
    def testSparseAdjacency(self):
        mat = sparse.coo_matrix(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 2, 0], [0, 1, 0, 0]],
        )

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat)
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (2, 2), (3, 1)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2), (2, 2)])

        # ADJ MAX
        g = Graph.Adjacency(mat, mode="max")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2), (2, 2), (1, 3)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2), (2, 2), (1, 3)])

        # ADJ UPPER
        g = Graph.Adjacency(mat, mode="upper")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2), (2, 2)])

    @unittest.skipIf(
        (sparse is None) or (np is None), "test case depends on NumPy/SciPy"
    )
    def testSparseAdjacencyLoopHandling(self):
        mat = sparse.coo_matrix(
            [[0, 1, 1, 0], [1, 0, 0, 0], [0, 0, 2, 0], [0, 1, 0, 0]],
        )

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat, loops=False)
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (3, 1)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min", loops=False)
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower", loops=False)
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (1, 3)])

        # ADJ_DIRECTED (default)
        g = Graph.Adjacency(mat, loops="twice")
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])

        # ADJ MAX
        g = Graph.Adjacency(mat, mode="max", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2), (1, 3)])

        # ADJ MIN
        g = Graph.Adjacency(mat, mode="min", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2)])

        # ADJ LOWER
        g = Graph.Adjacency(mat, mode="lower", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2), (1, 3)])

        # ADJ UPPER
        g = Graph.Adjacency(mat, mode="upper", loops="twice")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2)])

    def testWeightedAdjacency(self):
        mat = [[0, 1, 2, 0], [2, 0, 0, 0], [0, 0, 2.5, 0], [0, 1, 0, 0]]

        g = Graph.Weighted_Adjacency(mat, attr="w0")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 2.5, 1])

        g = Graph.Weighted_Adjacency(mat, mode="plus")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 3), (2, 2)])
        self.assertTrue(g.es["weight"] == [3, 2, 1, 2.5])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops=False)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops="twice")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1.25, 1])

    @unittest.skipIf(np is None, "test case depends on NumPy")
    def testWeightedAdjacencyNumPy(self):
        mat = np.array(
            [[0, 1, 2, 0], [2, 0, 0, 0], [0, 0, 2.5, 0], [0, 1, 0, 0]],
        )

        g = Graph.Weighted_Adjacency(mat, attr="w0")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 2.5, 1])

        g = Graph.Weighted_Adjacency(mat, mode="plus")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 3), (2, 2)])
        self.assertTrue(g.es["weight"] == [3, 2, 1, 2.5])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops=False)
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops="twice")
        el = g.get_edgelist()
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1.25, 1])

    @unittest.skipIf(
        (sparse is None) or (np is None), "test case depends on NumPy/SciPy"
    )
    def testSparseWeightedAdjacency(self):
        mat = sparse.coo_matrix(
            [[0, 1, 2, 0], [2, 0, 0, 0], [0, 0, 2.5, 0], [0, 1, 0, 0]]
        )

        g = Graph.Weighted_Adjacency(mat, attr="w0")
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 2.5, 1])

        g = Graph.Weighted_Adjacency(mat, mode="plus")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (2, 2), (1, 3)])
        self.assertTrue(g.es["weight"] == [3, 2, 2.5, 1])

        g = Graph.Weighted_Adjacency(mat, mode="min")
        el = g.get_edgelist()
        self.assertFalse(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (2, 2)])
        self.assertTrue(g.es["weight"] == [1, 2.5])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops=False)
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1])

        g = Graph.Weighted_Adjacency(mat, attr="w0", loops="twice")
        el = g.get_edgelist()
        self.assertTrue(g.is_directed())
        self.assertEqual(4, g.vcount())
        self.assertTrue(el == [(0, 1), (0, 2), (1, 0), (2, 2), (3, 1)])
        self.assertTrue(g.es["w0"] == [1, 2, 2, 1.25, 1])

    @unittest.skipIf((np is None) or (pd is None), "test case depends on NumPy/Pandas")
    def testDataFrame(self):
        edges = pd.DataFrame(
            [["C", "A", 0.4], ["A", "B", 0.1]], columns=[0, 1, "weight"]
        )
        g = Graph.DataFrame(edges, directed=False, use_vids=False)
        self.assertTrue(g.es["weight"] == [0.4, 0.1])

        vertices = pd.DataFrame(
            [["A", "blue"], ["B", "yellow"], ["C", "blue"]], columns=[0, "color"]
        )
        g = Graph.DataFrame(edges, directed=True, vertices=vertices, use_vids=False)
        self.assertTrue(g.vs["name"] == ["A", "B", "C"])
        self.assertTrue(g.vs["color"] == ["blue", "yellow", "blue"])
        self.assertTrue(g.es["weight"] == [0.4, 0.1])

        # Issue #347
        edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
        vertices = pd.DataFrame(
            {"node": [1, 2, 3, 4, 5, 6], "label": ["1", "2", "3", "4", "5", "6"]}
        )[["node", "label"]]
        g = Graph.DataFrame(edges, directed=True, vertices=vertices, use_vids=False)
        self.assertTrue(g.vs["name"] == [1, 2, 3, 4, 5, 6])
        self.assertTrue(g.vs["label"] == ["1", "2", "3", "4", "5", "6"])

        # Vertex names
        edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
        g = Graph.DataFrame(edges, use_vids=False)
        self.assertTrue(g.vcount() == 6)

        # Vertex ids
        edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
        g = Graph.DataFrame(edges)
        self.assertTrue(g.vcount() == 7)

        # Graph clone
        g = Graph.Full(n=100, directed=True, loops=True)
        g.vs["name"] = [f"v{i}" for i in range(g.vcount())]
        g.vs["x"] = [float(i) for i in range(g.vcount())]
        g.es["w"] = [1.0] * g.ecount()
        df_edges = g.get_edge_dataframe()
        df_vertices = g.get_vertex_dataframe()
        g_clone = Graph.DataFrame(df_edges, g.is_directed(), df_vertices)
        self.assertTrue(df_edges.equals(g_clone.get_edge_dataframe()))
        self.assertTrue(df_vertices.equals(g_clone.get_vertex_dataframe()))

        # pandas Int64 data type
        edges = pd.DataFrame(np.array([[0, 1], [1, 1], [1, 2]]), dtype="Int64")
        g = Graph.DataFrame(edges)
        self.assertTrue(g.vcount() == 3)

        # dataframe with both int data and str data
        edges = pd.DataFrame({"source": [1, 2, 2], "target": ["A", "B", "A"]})
        g = Graph.DataFrame(edges, use_vids=False)
        self.assertTrue(g.vs["name"] == [1, "A", 2, "B"])

        # Invalid input
        with self.assertRaisesRegex(ValueError, "two columns"):
            edges = pd.DataFrame({"source": [1, 2, 3]})
            Graph.DataFrame(edges)
        with self.assertRaisesRegex(ValueError, "one column"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            Graph.DataFrame(edges, vertices=pd.DataFrame())
        with self.assertRaisesRegex(TypeError, "integers"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]}).astype(str)
            Graph.DataFrame(edges)
        with self.assertRaisesRegex(ValueError, "negative"):
            edges = -pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            Graph.DataFrame(edges)
        with self.assertRaisesRegex(TypeError, "integers"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 3]}, index=["1", "2", "3"])
            Graph.DataFrame(edges, vertices=vertices)
        with self.assertRaisesRegex(ValueError, "negative"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 3]}, index=[-1, 2, 3])
            Graph.DataFrame(edges, vertices=vertices)
        with self.assertRaisesRegex(ValueError, "sequence"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 3]}, index=[1, 2, 4])
            Graph.DataFrame(edges, vertices=vertices)
        with self.assertRaisesRegex(TypeError, "integers"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame(
                {0: [1, 2, 3]},
                index=pd.MultiIndex.from_tuples([(1, 1), (2, 2), (3, 3)]),
            )
            Graph.DataFrame(edges, vertices=vertices)
        with self.assertRaisesRegex(ValueError, "unique"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 2]})
            Graph.DataFrame(edges, vertices=vertices, use_vids=False)
        with self.assertRaisesRegex(ValueError, "already contains"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 3], "name": [1, 2, 2]})
            Graph.DataFrame(edges, vertices=vertices, use_vids=False)
        with self.assertRaisesRegex(ValueError, "missing from"):
            edges = pd.DataFrame({"source": [1, 2, 3], "target": [4, 5, 6]})
            vertices = pd.DataFrame({0: [1, 2, 3]}, index=[0, 1, 2])
            Graph.DataFrame(edges, vertices=vertices)
        with self.assertRaisesRegex(ValueError, "null"):
            edges = pd.DataFrame(np.array([[0, 1], [1, np.nan], [1, 2]]), dtype="Int64")
            Graph.DataFrame(edges)


def suite():
    generator_suite = unittest.defaultTestLoader.loadTestsFromTestCase(GeneratorTests)
    return unittest.TestSuite([generator_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



import unittest

from igraph import Graph, InternalError, Layout


class GameTests(unittest.TestCase):
    def testGRG(self):
        g = Graph.GRG(50, 0.2)
        self.assertTrue(isinstance(g, Graph))
        g = Graph.GRG(50, 0.2, True)
        self.assertTrue(isinstance(g, Graph))
        self.assertTrue("x" in g.vertex_attributes())
        self.assertTrue("y" in g.vertex_attributes())
        self.assertTrue(isinstance(Layout(list(zip(g.vs["x"], g.vs["y"]))), Layout))

    def testForestFire(self):
        g = Graph.Forest_Fire(100, 0.1)
        self.assertTrue(isinstance(g, Graph) and g.is_directed() is False)
        g = Graph.Forest_Fire(100, 0.1, directed=True)
        self.assertTrue(isinstance(g, Graph) and g.is_directed() is True)

    def testRecentDegree(self):
        g = Graph.Recent_Degree(100, 5, 10)
        self.assertTrue(isinstance(g, Graph))

    def testPreference(self):
        g = Graph.Preference(100, [1, 1], [[1, 0], [0, 1]])
        self.assertTrue(isinstance(g, Graph))
        self.assertEqual(len(g.connected_components()), 2)

        g = Graph.Preference(100, [1, 1], [[1, 0], [0, 1]], attribute="type")
        types = g.vs.get_attribute_values("type")
        self.assertTrue(min(types) == 0 and max(types) == 1)

    def testAsymmetricPreference(self):
        g = Graph.Asymmetric_Preference(100, [[0, 1], [1, 0]], [[0, 1], [1, 0]])
        self.assertTrue(isinstance(g, Graph))
        self.assertEqual(len(g.connected_components()), 2)

        g = Graph.Asymmetric_Preference(
            100, [[0, 1], [1, 0]], [[1, 0], [0, 1]], attribute="type"
        )
        types = g.vs.get_attribute_values("type")
        types1 = [i[0] for i in types]
        types2 = [i[1] for i in types]
        self.assertTrue(
            min(types1) == 0
            and max(types1) == 1
            and min(types2) == 0
            and max(types2) == 1
        )

        g = Graph.Asymmetric_Preference(100, [[0, 1], [1, 0]], [[1, 0], [0, 1]])
        self.assertTrue(isinstance(g, Graph))
        self.assertEqual(len(g.connected_components()), 1)

    def testTreeGame(self):
        # Prufer algorithm
        g = Graph.Tree_Game(10, False, "Prufer")
        self.assertTrue(isinstance(g, Graph) and g.vcount() == 10)
        self.assertFalse(g.is_directed())
        self.assertTrue(g.is_tree())

        # Prufer with directed (should fail)
        self.assertRaises(InternalError, Graph.Tree_Game, 10, True, "Prufer")

        # LERW algorithm
        g = Graph.Tree_Game(10, False, "lerw")
        self.assertTrue(isinstance(g, Graph) and g.vcount() == 10)
        self.assertFalse(g.is_directed())
        self.assertTrue(g.is_tree())

        # Omitting the algorithm should default to LERW
        g = Graph.Tree_Game(10, directed=True)
        self.assertTrue(isinstance(g, Graph) and g.vcount() == 10)
        self.assertTrue(g.is_directed())
        self.assertTrue(g.is_tree())

        # Omitting the directed argument should use undirected graphs
        g = Graph.Tree_Game(42, method="Prufer")
        self.assertTrue(isinstance(g, Graph) and g.vcount() == 42)
        self.assertFalse(g.is_directed())
        self.assertTrue(g.is_tree())

    def testWattsStrogatz(self):
        g = Graph.Watts_Strogatz(1, 20, 1, 0.2)
        self.assertTrue(isinstance(g, Graph) and g.vcount() == 20 and g.ecount() == 20)

    def testRandomBipartiteNP(self):
        # Test np mode, undirected
        g = Graph.Random_Bipartite(10, 20, p=0.25)
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertFalse(g.is_directed())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])

        # Test np mode, directed, "out"
        g = Graph.Random_Bipartite(10, 20, p=0.25, directed=True, neimode="out")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])
        self.assertTrue(all(g.vs[e.tuple]["type"] == [False, True] for e in g.es))

        # Test np mode, directed, "in"
        g = Graph.Random_Bipartite(10, 20, p=0.25, directed=True, neimode="in")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])
        self.assertTrue(all(g.vs[e.tuple]["type"] == [True, False] for e in g.es))

        # Test np mode, directed, "all"
        g = Graph.Random_Bipartite(10, 20, p=0.25, directed=True, neimode="all")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])

    def testRandomBipartiteNM(self):
        # Test np mode, undirected
        g = Graph.Random_Bipartite(10, 20, m=50)
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertFalse(g.is_directed())
        self.assertEqual(50, g.ecount())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])

        # Test np mode, directed, "out"
        g = Graph.Random_Bipartite(10, 20, m=50, directed=True, neimode="out")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual(50, g.ecount())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])
        self.assertTrue(all(g.vs[e.tuple]["type"] == [False, True] for e in g.es))

        # Test np mode, directed, "in"
        g = Graph.Random_Bipartite(10, 20, m=50, directed=True, neimode="in")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual(50, g.ecount())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])
        self.assertTrue(all(g.vs[e.tuple]["type"] == [True, False] for e in g.es))

        # Test np mode, directed, "all"
        g = Graph.Random_Bipartite(10, 20, m=50, directed=True, neimode="all")
        self.assertTrue(g.is_simple())
        self.assertTrue(g.is_bipartite())
        self.assertTrue(g.is_directed())
        self.assertEqual(50, g.ecount())
        self.assertEqual([False] * 10 + [True] * 20, g.vs["type"])

    def testRewire(self):
        # Undirected graph
        g = Graph.GRG(25, 0.4)
        degrees = g.degree()

        # Rewiring without loops
        g.rewire(10000)
        self.assertEqual(degrees, g.degree())
        self.assertTrue(g.is_simple())

        # Rewiring with loops (1)
        g.rewire(10000, mode="loops")
        self.assertEqual(degrees, g.degree())
        self.assertFalse(any(g.is_multiple()))

        # Rewiring with loops (2)
        g = Graph.Full(4)
        g[1, 3] = 0
        degrees = g.degree()
        g.rewire(100, mode="loops")
        self.assertEqual(degrees, g.degree())
        self.assertFalse(any(g.is_multiple()))

        # Directed graph
        g = Graph.GRG(25, 0.4)
        g.to_directed("mutual")
        indeg, outdeg = g.indegree(), g.outdegree()
        g.rewire(10000)
        self.assertEqual(indeg, g.indegree())
        self.assertEqual(outdeg, g.outdegree())
        self.assertTrue(g.is_simple())

        # Directed graph with loops
        g.rewire(10000, mode="loops")
        self.assertEqual(indeg, g.indegree())
        self.assertEqual(outdeg, g.outdegree())
        self.assertFalse(any(g.is_multiple()))


def suite():
    game_suite = unittest.defaultTestLoader.loadTestsFromTestCase(GameTests)
    return unittest.TestSuite([game_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
.. _tutorials-delaunay-triangulation:

======================
Delaunay Triangulation
======================

This example demonstrates how to calculate the `Delaunay triangulation <https://en.wikipedia.org/wiki/Delaunay_triangulation>`_ of an input graph. We start by generating a set of points on a 2D grid using random ``numpy`` arrays and a graph with those vertex coordinates and no edges.

"""
import numpy as np
from scipy.spatial import Delaunay
import igraph as ig
import matplotlib.pyplot as plt


# %%
# We start by generating a random graph in the 2D unit cube, fixing the random
# seed to ensure reproducibility.
np.random.seed(0)
x, y = np.random.rand(2, 30)
g = ig.Graph(30)
g.vs['x'] = x
g.vs['y'] = y

# %%
# Because we already set the `x` and `y` vertex attributes, we can use
# :meth:`igraph.Graph.layout_auto` to wrap them into a :class:`igraph.Layout`
# object.
layout = g.layout_auto()

# %%
# Now we can calculate the delaunay triangulation using `scipy`'s :class:`scipy:scipy.spatial.Delaunay` class:
delaunay = Delaunay(layout.coords)

# %%
# Given the triangulation, we can add the edges into the original graph:
for tri in delaunay.simplices:
    g.add_edges([
        (tri[0], tri[1]),
        (tri[1], tri[2]),
        (tri[0], tri[2]),
    ])

# %%
# Because adjacent triangles share an edge/side, the graph now contains some
# edges more than once. It's useful to simplify the graph to get rid of those
# multiedges, keeping each side only once:
g.simplify()

# %%
# Finally, plotting the graph gives a good idea of what the triangulation looks
# like:
fig, ax = plt.subplots()
ig.plot(
    g,
    layout=layout,
    target=ax,
    vertex_size=4,
    vertex_color="lightblue",
    edge_width=0.8
)
plt.show()

# %%
# Alternative plotting style
# --------------------------
# We can use :doc:`matplotlib <matplotlib:index>` to plot the faces of the
# triangles instead of the edges. First, we create a hue gradient for the
# triangle faces:
palette = ig.GradientPalette("midnightblue", "lightblue", 100)

# %%
# Then we can "plot" the triangles using
# :class:`matplotlib:matplotlib.patches.Polygon` and the graph using
# :func:`igraph.plot() <igraph.drawing.plot>`:
fig, ax = plt.subplots()
for tri in delaunay.simplices:
    # get the points of the triangle
    tri_points = [delaunay.points[tri[i]] for i in range(3)]

    # calculate the vertical center of the triangle
    center = (tri_points[0][1] + tri_points[1][1] + tri_points[2][1]) / 3

    # draw triangle onto axes
    poly = plt.Polygon(tri_points, color=palette.get(int(center * 100)))
    ax.add_patch(poly)

ig.plot(
    g,
    layout=layout,
    target=ax,
    vertex_size=0,
    edge_width=0.2,
    edge_color="white",
)
ax.set(xlim=(0, 1), ylim=(0, 1))
plt.show()



#! /usr/bin/env python3

#   IGraph library
#   Copyright (C) 2005-2021  The igraph development team
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc.,  51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301 USA
#
###################################################################

"""DocBook XML generator for igraph.

The generator parses one or more input files for documentation chunks
(embedded in the source code as Doxygen-style comments), and processes
them with a set of regex-based rules. The processed chunks are then
substituted into a template file containing <!-- doxrox-include -->
directives.

When a template file is not provided, the generator will read the input
files, process them with the ruleset and save a dictionary mapping chunk
names to the corresponding processed chunks into a Python pickle. This
can be used to speed up the processing of multiple input files as you can
generate the chunks once and then re-use them for multiple input files.
"""

import os
import re
import sys

from argparse import ArgumentParser
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from fnmatch import fnmatch
from hashlib import sha1
from operator import itemgetter
from pathlib import Path
from pickle import dump, load
from time import time
from typing import Any, Callable, Dict, Iterator, List, Optional, Pattern

#: Constant indicating the start of a comment that doxrox.py will process
DOXHEAD: str = r"/\*\*"

#: Stores whether we want verbose output
verbose: bool = False


def fatal(message: str, code: int = 1):
    """Prints an error message and exits the program with the given error code."""
    print(message, file=sys.stderr)
    sys.exit(code)


#########################################################################
# The main function
#########################################################################


def main():
    """Main entry point of the script."""

    global verbose

    # get command line arguments
    parser = create_argument_parser()
    arguments = parser.parse_args()

    outputfile: str = arguments.output_file
    inputs: List[str] = arguments.inputs

    verbose = arguments.verbose

    if (
        arguments.template_file in inputs
        or arguments.rules_file in inputs
        or outputfile in inputs
    ):
        fatal("Special file is also used as an input file", 2)

    # open the cache file if needed
    cache = ChunkCache(arguments.cache_file) if arguments.cache_file else None

    # get all regular expressions
    rules: List[Rule]
    if arguments.rules_file:
        with operation("Reading regular expressions...") as op:
            rules = read_regex_rules_file(arguments.rules_file)
            op("{0} rules read".format(len(rules)))
    else:
        rules = []

    # parse all input files and extract chunks, apply rules
    if arguments.chunk_file:
        with operation("Reading pickled chunks...") as op:
            try:
                with open(arguments.chunk_file, "rb") as f:
                    all_chunks = load(f)
            except IOError:
                fatal("Error reading chunk file: " + arguments.chunk_file, 9)
            op("{0} chunks read".format(len(all_chunks)))
    else:
        all_chunks = {}

    rule_timings = defaultdict(list)

    for ifile in inputs:
        with operation("Parsing input file {0}...".format(ifile)) as op:
            try:
                with open(ifile, "r") as f:
                    contents = f.read()
            except IOError:
                fatal("Error reading input file: " + ifile, 3)

            if cache:
                key = cache.key_of(contents)
                chunks = cache.get(key)
            else:
                key, chunks = None, None

            if chunks is not None:
                op("{0} chunks read from cache".format(len(chunks)))
            else:
                chunks = collect_chunks_from_input_file(
                    ifile, contents, rules, rule_timings
                )
                op("{0} chunks parsed".format(len(chunks)))
                if key and cache:
                    cache.put(key, chunks)

            for name, chunk in chunks.items():
                if name in all_chunks:
                    fatal(
                        "Multiple files provide chunks for {0!r}".format(name), code=4
                    )
                all_chunks[name] = chunk

    if arguments.timing_stats and rule_timings:
        rule_timings = {name: sum(dts) / len(dts) for name, dts in rule_timings.items()}
        for name, dt in sorted(rule_timings.items(), key=itemgetter(1), reverse=True):
            print("{0}: {1:.3f}us".format(name, dt))
        print("======")

    if cache:
        cache.close()

    if arguments.template_file:
        # substitute the template file
        with operation("Reading template file..."):
            try:
                with open(arguments.template_file, "r") as tfile:
                    tstring = tfile.read()
            except IOError:
                fatal("Error reading the template file: " + arguments.template_file, 7)

        with operation("Substituting template file..."):
            chunk_iterator = re.finditer(
                r"<!--\s*doxrox-include\s+(\w+)\s*-->", tstring
            )
            outstring = []
            last = 0
            for match in chunk_iterator:
                try:
                    chunk = all_chunks[match.group(1)]
                except KeyError:
                    fatal("Chunk not found: {0}".format(match.group(1)), code=4)
                outstring.append(tstring[last : match.start()])
                outstring.append(chunk)
                last = match.end()
            outstring.append(tstring[last:])
            outstring = "".join(outstring)

        # write output file
        with operation("Writing output file..."):
            try:
                with open(outputfile, "w") as ofile:
                    ofile.write(outstring)
            except IOError:
                fatal("Error writing output file:" + outputfile, 8)
    else:
        # no template file given so just save the chunks as a pickle into the
        # output file
        with operation("Writing output file..."):
            try:
                with open(outputfile, "wb") as ofile:
                    dump(all_chunks, ofile)
            except IOError:
                fatal("Error writing output file:" + outputfile, 5)


#########################################################################
# Argument parser
#########################################################################


def create_argument_parser() -> ArgumentParser:
    """Creates the command line argument parser that the script uses."""
    parser = ArgumentParser(description=(sys.modules[__name__].__doc__ or "").strip())

    parser.add_argument(
        "--cache",
        metavar="FILE",
        dest="cache_file",
        help="optional cache file to store chunks from already processed files",
    )
    parser.add_argument(
        "-t",
        "--template",
        metavar="FILE",
        dest="template_file",
        help="template file to process",
    )
    parser.add_argument(
        "-e",
        "--rules",
        metavar="FILE",
        dest="rules_file",
        help="file containing matching and replacement rules",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        dest="output_file",
        required=True,
        help="name of the output file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="enable verbose output",
    )
    parser.add_argument(
        "--chunks",
        dest="chunk_file",
        metavar="FILE",
        help="name of a previously saved chunk file",
    )
    parser.add_argument(
        "--timing-stats",
        dest="timing_stats",
        action="store_true",
        default=False,
        help="print the average time it takes to process regex rules from the rules file",
    )
    parser.add_argument(
        "inputs", metavar="INPUT", nargs="*", help="input files to process"
    )

    return parser


#################
# classes and functions to read the regular expression rules
#################


class RuleType(Enum):
    REPLACE = "replace"
    RUN = "run"


@dataclass
class Rule:
    regex: Pattern[str]
    """The regular expression that the rule will attempt to match."""

    replacement: str
    """The replacement string for the match, or the code to execute on the
    match.
    """

    type: RuleType
    """Type of the rule"""

    name: Optional[str]
    """Name of the rule, for debugging purposes."""

    glob: Optional[str] = None
    """Optional glob pattern that specifies which input files the rule
    applies to.
    """

    def applies_to_filename(self, filename: str) -> bool:
        """Returns whether the rule applies to files with the given name."""
        if self.glob:
            return fnmatch(filename, self.glob)
        else:
            return True


def read_regex_rules_file(filename) -> List[Rule]:
    """Parses the file containing the regex-based rules that we use to chop
    up the input source files into chunks that can later be fed into a
    DocBook document.

    Parameters:
        filename: name of the input file

    Returns:
        the rules that were parsed from the input file
    """

    rules: List[Rule] = []

    def parse_error(lineno):
        """Helper function to indicate a parse error at the given line."""
        fatal(
            "Parse error in regex file ({0}), line {1}".format(filename, lineno), code=4
        )

    def store(
        rule: List[str],
        replacement: List[str],
        rule_name: Optional[str],
        rule_type: RuleType,
        glob: Optional[str],
    ) -> None:
        """Helper function to append the current rule to the result."""
        regex = re.compile("".join(rule), re.VERBOSE | re.MULTILINE | re.DOTALL)
        replacement_str = "".join(replacement)[:-1]
        rules.append(Rule(regex, replacement_str, rule_type, rule_name, glob))

    mode = "empty"
    regex, replacement = [], []
    rule_name: Optional[str] = None
    rule_type: Optional[RuleType] = None
    glob: Optional[str] = None

    try:
        with open(filename, "r") as f:
            for lineno, line in enumerate(f, 1):
                if line.startswith("REPLACE"):
                    # a new pattern block starts
                    if mode not in ("empty", "with"):
                        parse_error(lineno)
                    else:
                        if regex and rule_type:
                            store(regex, replacement, rule_name, rule_type, glob)
                        regex.clear()
                        replacement.clear()
                        mode = "replace"

                        match = re.match(
                            r"^REPLACE( IN (?P<glob>[^\s]+))?\s+-+\s+(?P<name>.*)\s+-",
                            line,
                        )
                        rule_name = match.group("name") if match else None
                        glob = match.group("glob") if match else None

                elif line.startswith("WITH") or line.startswith("RUN"):
                    # the second half of the pattern block starts
                    if mode != "replace":
                        parse_error(lineno)
                    else:
                        mode = "with"
                    rule_type = (
                        RuleType.REPLACE if line.startswith("WITH") else RuleType.RUN
                    )

                elif re.match(r"^\s*$", line):
                    # empty line, do nothing
                    pass

                else:
                    # normal line, append
                    if mode == "replace":
                        regex.append(line)
                    elif mode == "with":
                        replacement.append(line)
                    else:
                        parse_error(lineno)

            if regex != "" and rule_type:
                store(regex, replacement, rule_name, rule_type, glob)

    except IOError:
        fatal("Error reading regex file: " + filename, code=4)

    return rules


#################
# parse an input file string
#################
def collect_chunks_from_input_file(
    path: str, strinput: str, rules: List[Rule], rule_timings
) -> Dict[str, str]:
    result: Dict[str, str] = {}

    # split the file
    chunks = re.split(DOXHEAD, strinput)
    chunks = chunks[1:]

    # get the filename part of the path
    filename = os.path.basename(path)

    # apply all rules to the chunks
    for chunk in chunks:
        name: Optional[str] = None

        for rule in rules:
            start = time()

            if not name and "name" in rule.regex.groupindex:
                # The regex might provide us with a chunk name so try figuring
                # out what the "name" group might match to
                matched = rule.regex.search(chunk)
                if matched:
                    try:
                        name = matched.group("name")
                    except IndexError:
                        name = ""

            if rule.applies_to_filename(filename):
                if rule.type is RuleType.REPLACE:
                    # This is a simple regex replacement rule
                    try:
                        chunk = rule.regex.sub(rule.replacement, chunk)
                    except IndexError:
                        print("Index error:" + chunk[0:60] + "...")
                        print("Pattern:\n" + rule.regex.pattern)
                        print("Current state:" + chunk[0:60] + "...")
                        fatal("Parsing error", code=6)
                elif rule.type is RuleType.RUN:
                    # This is a piece of Python code that has to be executed on
                    # the part that matched
                    matched = rule.regex.search(chunk)
                    if matched:
                        exec(rule.replacement)
                else:
                    fatal("Invalid rule type: {0!r}".format(rule.type), code=6)

            rule_timings[rule.name].append((time() - start) * 1000000)

        if not name:
            # print("Chunk without a name ignored:" + ch[0:60] + "...")
            continue

        result[name] = chunk.strip()

    return result


@contextmanager
def operation(message: str) -> Iterator[Callable[[Any], None]]:
    """Helper function to show progress messages for a potentially long-running
    operation in verbose mode.

    Parameters:
        message (str): the message to show
    """
    global verbose

    if verbose:
        print(message, end="")

    result = [None]

    def set_result(obj: Any) -> None:
        result[0] = obj

    success = False
    try:
        yield set_result
        success = True
    finally:
        if verbose and success:
            if result[0] is None:
                print(" done.")
            else:
                print(" done, {0}.".format(result[0]))


class ChunkCache:
    """Simple on-disk cache that stores SHA256 hashes of files along with the
    DocBook documentation chunks that were parsed from them.
    """

    _data: Optional[Dict[str, Dict[str, str]]]
    _dirty: bool
    _path: Path

    def __init__(self, filename: str, hash=sha1):
        """Constructor.

        Parameters:
            filename: name of the file on the disk where the cache resides
            hash: the hash function to use
        """
        self._data = None
        self._dirty = False
        self._hash = hash
        self._path = Path(filename)

    def _load(self) -> None:
        """Populates the in-memory copy of the cache from the disk."""
        if self._path.exists():
            try:
                with self._path.open("rb") as fp:
                    self._data = load(fp)
            except (IOError, EOFError):
                # cache corrupted
                self._data = {}
        else:
            self._data = {}
        self._dirty = False

    def close(self) -> None:
        """Closes the cache and flushes its contents back to the disk if it
        changed recently.
        """
        if self._dirty:
            self.flush()

    def flush(self) -> None:
        """Flushes the contents of the cache back to the disk."""
        with self._path.open("wb") as fp:
            dump(self._data, fp)
        self._dirty = False

    def get(self, key: str) -> Optional[Dict[str, str]]:
        """Returns the chunks associated to the file with the given key, or
        `None` if the key is not in the cache.
        """
        if self._data is None:
            self._load()

        assert self._data is not None
        return self._data.get(key)

    def key_of(self, contents, encoding: str = "utf-8") -> str:
        """Returns the hash key corresponding to the file with the given
        contents.
        """
        if not isinstance(contents, bytes):
            contents = contents.encode(encoding)

        key = self._hash()
        key.update(contents)
        return key.hexdigest()

    def put(self, key: str, chunks: Dict[str, str]) -> None:
        """Stores some chunks associated to the file with the given key."""
        assert self._data is not None
        self._data[key] = chunks
        self._dirty = True


if __name__ == "__main__":
    main()



"""
.. _tutorials-spanning-trees:

==============
Spanning Trees
==============

This example shows how to generate a spanning tree from an input graph using :meth:`igraph.Graph.spanning_tree`. For the related idea of finding a *minimum spanning tree*, see :ref:`tutorials-minimum-spanning-trees`.
"""
import igraph as ig
import matplotlib.pyplot as plt
import random

# %%
# First we create a two-dimensional, 6 by 6 lattice graph:
g = ig.Graph.Lattice([6, 6], circular=False)

# %%
# We can compute the 2D layout of the graph:
layout = g.layout("grid")

# %%
# To spice things up a little, we rearrange the vertex ids and compute a new
# layout. While not terribly useful in this context, it does make for a more
# interesting-looking spanning tree ;-)
random.seed(0)
permutation = list(range(g.vcount()))
random.shuffle(permutation)
g = g.permute_vertices(permutation)
new_layout = g.layout("grid")
for i in range(36):
    new_layout[permutation[i]] = layout[i]
layout = new_layout

# %%
# We can now generate a spanning tree:
spanning_tree = g.spanning_tree(weights=None, return_tree=False)

# %%
# Finally, we can plot the graph with a highlight color for the spanning tree.
# We follow the usual recipe: first we set a few aesthetic options and then we
# leverage :func:`igraph.plot() <igraph.drawing.plot>` and matplotlib for the
# heavy lifting:
g.es["color"] = "lightgray"
g.es[spanning_tree]["color"] = "midnightblue"
g.es["width"] = 0.5
g.es[spanning_tree]["width"] = 3.0

fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout=layout,
    vertex_color="lightblue",
    edge_width=g.es["width"]
)
plt.show()

# %%
# .. note::
#   To invert the y axis such that the root of the tree is on top of the plot,
#   you can call `ax.invert_yaxis()` before `plt.show()`.



def _construct_random_geometric_graph(cls, n, radius, torus=False):
    """Generates a random geometric graph.

    The algorithm drops the vertices randomly on the 2D unit square and
    connects them if they are closer to each other than the given radius.
    The coordinates of the vertices are stored in the vertex attributes C{x}
    and C{y}.

    @param n: The number of vertices in the graph
    @param radius: The given radius
    @param torus: This should be C{True} if we want to use a torus instead of a
      square.
    """
    result, xs, ys = cls._GRG(n, radius, torus)
    result.vs["x"] = xs
    result.vs["y"] = ys
    return result



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""
Shape drawing classes for igraph

Vertex shapes in igraph are usually referred to by short names like
C{"rect"} or C{"circle"}. This module contains the classes that
implement the actual drawing routines for these shapes, and a
resolver class that determines the appropriate shape drawer class
given the short name.

Classes that are derived from L{ShapeDrawer} in this module are
automatically registered by L{ShapeDrawerDirectory}. If you
implement a custom shape drawer, you must register it in
L{ShapeDrawerDirectory} manually if you wish to refer to it by a
name in the C{shape} attribute of vertices.
"""


__all__ = ("ShapeDrawerDirectory",)

from abc import ABCMeta, abstractmethod
from math import atan2, copysign, cos, pi, sin
import sys

from igraph.drawing.matplotlib.utils import find_matplotlib

mpl, plt = find_matplotlib()


class ShapeDrawer(metaclass=ABCMeta):
    """Static class, the ancestor of all vertex shape drawer classes.

    Custom shapes must implement at least the C{draw_path} method of the class.
    The method I{must not} stroke or fill, it should just set up the current
    Cairo path appropriately."""

    @staticmethod
    @abstractmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws the path of the shape on the given Cairo context, without
        stroking or filling it.

        This method must be overridden in derived classes implementing custom shapes
        and declared as a static method using C{staticmethod(...)}.

        @param ctx: the context to draw on
        @param center_x: the X coordinate of the center of the object
        @param center_y: the Y coordinate of the center of the object
        @param width: the width of the object
        @param height: the height of the object. If C{None}, equals to the width.
        """
        raise NotImplementedError

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the shape centered at (center_x, center_y)
        intersects with a line drawn from (source_x, source_y) to
        (center_x, center_y).

        Can be overridden in derived classes. Must always be defined as a static
        method using C{staticmethod(...)}

        @param width: the width of the shape
        @param height: the height of the shape. If C{None}, defaults to the width
        @return: the intersection point (the closest to (source_x, source_y) if
            there are more than one) or (center_x, center_y) if there is no
            intersection
        """
        return center_x, center_y


class NullDrawer(ShapeDrawer):
    """Static drawer class which draws nothing.

    This class is used for graph vertices with unknown shapes"""

    names = ["null", "none", "empty", "hidden", ""]

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None):
        """Draws nothing."""
        pass


class RectangleDrawer(ShapeDrawer):
    """Static class which draws rectangular vertices"""

    names = "rectangle rect rectangular square box s"

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws a rectangle-shaped path on the Cairo context without stroking
        or filling it.
        @see: ShapeDrawer.draw_path"""
        height = height or width
        if hasattr(plt, "Axes") and isinstance(ctx, plt.Axes):
            return mpl.patches.Rectangle(
                (-width / 2, -height / 2),
                width,
                height,
                clip_on=True,
                **kwargs,
            )
        else:
            ctx.rectangle(center_x - width / 2, center_y - height / 2, width, height)

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the rectangle centered at (center_x, center_y)
        having the given width and height intersects with a line drawn from
        (source_x, source_y) to (center_x, center_y).

        @see: ShapeDrawer.intersection_point"""
        height = height or width
        delta_x, delta_y = center_x - source_x, center_y - source_y

        if delta_x == 0 and delta_y == 0:
            return center_x, center_y

        if delta_y > 0 and delta_x <= delta_y and delta_x >= -delta_y:
            # this is the top edge
            ry = center_y - height / 2
            ratio = (height / 2) / delta_y
            return center_x - ratio * delta_x, ry

        if delta_y < 0 and delta_x <= -delta_y and delta_x >= delta_y:
            # this is the bottom edge
            ry = center_y + height / 2
            ratio = (height / 2) / -delta_y
            return center_x - ratio * delta_x, ry

        if delta_x > 0 and delta_y <= delta_x and delta_y >= -delta_x:
            # this is the left edge
            rx = center_x - width / 2
            ratio = (width / 2) / delta_x
            return rx, center_y - ratio * delta_y

        if delta_x < 0 and delta_y <= -delta_x and delta_y >= delta_x:
            # this is the right edge
            rx = center_x + width / 2
            ratio = (width / 2) / -delta_x
            return rx, center_y - ratio * delta_y

        if delta_x == 0:
            if delta_y > 0:
                return center_x, center_y - height / 2
            return center_x, center_y + height / 2

        if delta_y == 0:
            if delta_x > 0:
                return center_x - width / 2, center_y
            return center_x + width / 2, center_y


class CircleDrawer(ShapeDrawer):
    """Static class which draws circular vertices"""

    names = "circle circular o"

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws a circular path on the Cairo context without stroking or
        filling it.

        Height is ignored, it is the width that determines the diameter of the circle.

        @see: ShapeDrawer.draw_path"""
        if hasattr(plt, "Axes") and isinstance(ctx, plt.Axes):
            return mpl.patches.Circle(
                (0, 0),
                width / 2,
                clip_on=True,
                **kwargs,
            )
        else:
            ctx.arc(center_x, center_y, width / 2, 0, 2 * pi)

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the circle centered at (center_x, center_y)
        intersects with a line drawn from (source_x, source_y) to
        (center_x, center_y).

        @see: ShapeDrawer.intersection_point"""
        height = height or width
        angle = atan2(center_y - source_y, center_x - source_x)
        return center_x - width / 2 * cos(angle), center_y - height / 2 * sin(angle)


class UpTriangleDrawer(ShapeDrawer):
    """Static class which draws upright triangles"""

    names = "triangle triangle-up up-triangle arrow arrow-up up-arrow ^"

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws an upright triangle on the Cairo context without stroking or
        filling it.

        @see: ShapeDrawer.draw_path"""
        height = height or width
        if hasattr(plt, "Axes") and isinstance(ctx, plt.Axes):
            vertices = [
                [center_x - 0.5 * width, center_y - 0.333 * height],
                [center_x + 0.5 * width, center_y - 0.333 * height],
                [center_x, center_x + 0.667 * height],
            ]
            return mpl.patches.Polygon(
                vertices,
                closed=True,
                clip_on=True,
                **kwargs,
            )
        else:
            ctx.move_to(center_x - width / 2, center_y + height / 2)
            ctx.line_to(center_x, center_y - height / 2)
            ctx.line_to(center_x + width / 2, center_y + height / 2)
            ctx.close_path()

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the triangle centered at (center_x, center_y)
        intersects with a line drawn from (source_x, source_y) to
        (center_x, center_y).

        @see: ShapeDrawer.intersection_point"""
        # TODO: finish it properly
        height = height or width
        return center_x, center_y


class DownTriangleDrawer(ShapeDrawer):
    """Static class which draws triangles pointing down"""

    names = "down-triangle triangle-down arrow-down down-arrow v"

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws a triangle on the Cairo context without stroking or
        filling it.

        @see: ShapeDrawer.draw_path"""
        height = height or width
        if hasattr(plt, "Axes") and isinstance(ctx, plt.Axes):
            vertices = [
                [-0.5 * width, 0.333 * height],
                [0.5 * width, 0.333 * height],
                [0, -0.667 * height],
            ]
            return mpl.patches.Polygon(
                vertices,
                closed=True,
                clip_on=True,
                **kwargs,
            )

        else:
            ctx.move_to(center_x - width / 2, center_y - height / 2)
            ctx.line_to(center_x, center_y + height / 2)
            ctx.line_to(center_x + width / 2, center_y - height / 2)
            ctx.close_path()

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the triangle centered at (center_x, center_y)
        intersects with a line drawn from (source_x, source_y) to
        (center_x, center_y).

        @see: ShapeDrawer.intersection_point"""
        # TODO: finish it properly
        height = height or width
        return center_x, center_y


class DiamondDrawer(ShapeDrawer):
    """Static class which draws diamonds (i.e. rhombuses)"""

    names = "diamond rhombus d"

    @staticmethod
    def draw_path(ctx, center_x, center_y, width, height=None, **kwargs):
        """Draws a rhombus on the Cairo context without stroking or
        filling it.

        @see: ShapeDrawer.draw_path"""
        height = height or width
        if hasattr(plt, "Axes") and isinstance(ctx, plt.Axes):
            vertices = [
                [-0.5 * width, 0],
                [0, -0.5 * height],
                [0.5 * width, 0],
                [0, 0.5 * height],
            ]
            return mpl.patches.Polygon(
                vertices,
                closed=True,
                clip_on=True,
                **kwargs,
            )
        else:
            ctx.move_to(center_x - width / 2, center_y)
            ctx.line_to(center_x, center_y + height / 2)
            ctx.line_to(center_x + width / 2, center_y)
            ctx.line_to(center_x, center_y - height / 2)
            ctx.close_path()

    @staticmethod
    def intersection_point(center_x, center_y, source_x, source_y, width, height=None):
        """Determines where the rhombus centered at (center_x, center_y)
        intersects with a line drawn from (source_x, source_y) to
        (center_x, center_y).

        @see: ShapeDrawer.intersection_point"""
        height = height or width

        if height == 0 and width == 0:
            return center_x, center_y

        delta_x, delta_y = source_x - center_x, source_y - center_y

        # Treat edge case when delta_x = 0
        if delta_x == 0:
            if delta_y == 0:
                return center_x, center_y
            else:
                return center_x, center_y + copysign(height / 2, delta_y)

        width = copysign(width, delta_x)
        height = copysign(height, delta_y)

        f = height / (height + width * delta_y / delta_x)
        return center_x + f * width / 2, center_y + (1 - f) * height / 2


#####################################################################


class ShapeDrawerDirectory:
    """Static class that resolves shape names to their corresponding
    shape drawer classes.

    Classes that are derived from L{ShapeDrawer} in this module are
    automatically registered by L{ShapeDrawerDirectory} when the module
    is loaded for the first time.
    """

    known_shapes = {}

    @classmethod
    def register(cls, drawer_class):
        """Registers the given shape drawer class under the given names.

        @param drawer_class: the shape drawer class to be registered
        """
        names = drawer_class.names
        if isinstance(names, str):
            names = names.split()

        for name in names:
            cls.known_shapes[name] = drawer_class

    @classmethod
    def register_namespace(cls, namespace):
        """Registers all L{ShapeDrawer} classes in the given namespace

        @param namespace: a Python dict mapping names to Python objects."""
        for name, value in namespace.items():
            if name.startswith("__"):
                continue
            if isinstance(value, type):
                if issubclass(value, ShapeDrawer) and value != ShapeDrawer:
                    cls.register(value)

    @classmethod
    def resolve(cls, shape):
        """Given a shape name, returns the corresponding shape drawer class

        @param shape: the name of the shape
        @return: the corresponding shape drawer class

        @raise ValueError: if the shape is unknown
        """
        try:
            return cls.known_shapes[shape]
        except KeyError:
            raise ValueError("unknown shape: %s" % shape) from None

    @classmethod
    def resolve_default(cls, shape, default=NullDrawer):
        """Given a shape name, returns the corresponding shape drawer class
        or the given default shape drawer if the shape name is unknown.

        @param shape: the name of the shape
        @param default: the default shape drawer to return when the shape
          is unknown
        @return: the shape drawer class corresponding to the given name or
          the default shape drawer class if the name is unknown
        """
        return cls.known_shapes.get(shape, default)


ShapeDrawerDirectory.register_namespace(sys.modules[__name__].__dict__)



"""
.. _tutorials-random:

=================
Erdős-Rényi Graph
=================

This example demonstrates how to generate `Erdős–Rényi graphs <https://en.wikipedia.org/wiki/Erd%C5%91s%E2%80%93R%C3%A9nyi_model>`_ using :meth:`igraph.GraphBase.Erdos_Renyi`. There are two variants of graphs:

- ``Erdos_Renyi(n, p)`` will generate a graph from the so-called :math:`G(n,p)` model where each edge between any two pair of nodes has an independent probability ``p`` of existing.
- ``Erdos_Renyi(n, m)`` will pick a graph uniformly at random out of all graphs with ``n`` nodes and ``m`` edges. This is referred to as the :math:`G(n,m)` model.

We generate two graphs of each, so we can confirm that our graph generator is truly random.
"""
import igraph as ig
import matplotlib.pyplot as plt
import random

# %%
# First, we set a random seed for reproducibility
random.seed(0)

# %%
# Then, we generate two :math:`G(n,p)` Erdős–Rényi graphs with identical parameters:
g1 = ig.Graph.Erdos_Renyi(n=15, p=0.2, directed=False, loops=False)
g2 = ig.Graph.Erdos_Renyi(n=15, p=0.2, directed=False, loops=False)

# %%
# For comparison, we also generate two :math:`G(n,m)` Erdős–Rényi graphs with a fixed number
# of edges:
g3 = ig.Graph.Erdos_Renyi(n=20, m=35, directed=False, loops=False)
g4 = ig.Graph.Erdos_Renyi(n=20, m=35, directed=False, loops=False)

# %%
# We can print out summaries of each graph to verify their randomness
ig.summary(g1)
ig.summary(g2)
ig.summary(g3)
ig.summary(g4)

# IGRAPH U--- 15 18 --
# IGRAPH U--- 15 21 --
# IGRAPH U--- 20 35 --
# IGRAPH U--- 20 35 --

# %%
# Finally, we can plot the graphs to illustrate their structures and
# differences:
fig, axs = plt.subplots(2, 2)
# Probability
ig.plot(
    g1,
    target=axs[0, 0],
    layout="circle",
    vertex_color="lightblue"
)
ig.plot(
    g2,
    target=axs[0, 1],
    layout="circle",
    vertex_color="lightblue"
)
axs[0, 0].set_ylabel('Probability')
# N edges
ig.plot(
    g3,
    target=axs[1, 0],
    layout="circle",
    vertex_color="lightblue",
    vertex_size=15
)
ig.plot(
    g4,
    target=axs[1, 1],
    layout="circle",
    vertex_color="lightblue",
    vertex_size=15
)
axs[1, 0].set_ylabel('N. edges')
plt.show()



_format_mapping = {
    "ncol": ("Read_Ncol", "write_ncol"),
    "lgl": ("Read_Lgl", "write_lgl"),
    "graphdb": ("Read_GraphDB", None),
    "graphmlz": ("Read_GraphMLz", "write_graphmlz"),
    "graphml": ("Read_GraphML", "write_graphml"),
    "gml": ("Read_GML", "write_gml"),
    "dot": (None, "write_dot"),
    "graphviz": (None, "write_dot"),
    "net": ("Read_Pajek", "write_pajek"),
    "pajek": ("Read_Pajek", "write_pajek"),
    "dimacs": ("Read_DIMACS", "write_dimacs"),
    "adjacency": ("Read_Adjacency", "write_adjacency"),
    "adj": ("Read_Adjacency", "write_adjacency"),
    "edgelist": ("Read_Edgelist", "write_edgelist"),
    "edge": ("Read_Edgelist", "write_edgelist"),
    "edges": ("Read_Edgelist", "write_edgelist"),
    "pickle": ("Read_Pickle", "write_pickle"),
    "picklez": ("Read_Picklez", "write_picklez"),
    "svg": (None, "write_svg"),
    "gw": (None, "write_leda"),
    "leda": (None, "write_leda"),
    "lgr": (None, "write_leda"),
    "dl": ("Read_DL", None),
}



"""
.. _tutorials-cliques:

============
Cliques
============

This example shows how to compute and visualize cliques of a graph using :meth:`igraph.GraphBase.cliques`.

"""
import igraph as ig
import matplotlib.pyplot as plt

# %%
# First, let's create a graph, for instance the famous karate club graph:
g = ig.Graph.Famous('Zachary')

# %%
# Computing cliques can be done as follows:
cliques = g.cliques(4, 4)

# %%
# We can plot the result of the computation. To make things a little more
# interesting, we plot each clique highlighted in a separate axes:
fig, axs = plt.subplots(3, 4)
axs = axs.ravel()
for clique, ax in zip(cliques, axs):
    ig.plot(
        ig.VertexCover(g, [clique]),
        mark_groups=True, palette=ig.RainbowPalette(),
        vertex_size=5,
        edge_width=0.5,
        target=ax,
    )
plt.axis('off')
plt.show()


# %%
# Advanced: improving plotting style
# ----------------------------------
# If you want a little more style, you can color the vertices/edges within each
# clique to make them stand out:
fig, axs = plt.subplots(3, 4)
axs = axs.ravel()
for clique, ax in zip(cliques, axs):
    # Color vertices yellow/red based on whether they are in this clique
    g.vs['color'] = 'yellow'
    g.vs[clique]['color'] = 'red'

    # Color edges black/red based on whether they are in this clique
    clique_edges = g.es.select(_within=clique)
    g.es['color'] = 'black'
    clique_edges['color'] = 'red'
    # also increase thickness of clique edges
    g.es['width'] = 0.3
    clique_edges['width'] = 1

    ig.plot(
        ig.VertexCover(g, [clique]),
        mark_groups=True,
        palette=ig.RainbowPalette(),
        vertex_size=5,
        target=ax,
    )
plt.axis('off')
plt.show()



import os
import inspect

import igraph


# FIXME: there must be a better way to do this
auxiliary_imports = [
    ("typing", "*"),
    ("igraph.io.files", "_identify_format"),
    ("igraph.community", "_optimal_cluster_count_from_merges_and_modularity"),
]


def main():

    # Get instance and classmethods
    g = igraph.Graph()
    methods = inspect.getmembers(g, predicate=inspect.ismethod)

    # Get the source code for each method and replace the method name
    # in the signature
    methodsources = {}
    underscore_functions = set()
    for mname, method in methods:
        # Source code of the function that uses the method
        source = inspect.getsourcelines(method)[0]

        # Find function name for this modularized method
        fname = source[0][source[0].find("def ") + 4 : source[0].find("(")]

        # FIXME: this also swaps in methods that are already there. While
        # that should be fine, we could check

        # Make new source code, which is the same but with the name swapped
        newsource = [source[0].replace(fname, mname)] + source[1:]
        methodsources[mname] = newsource

        # Prepare to delete the import for underscore functions
        if fname.startswith("_"):
            underscore_functions.add(fname)

    newmodule = igraph.__file__ + ".new"
    with open(newmodule, "wt") as fout:
        # FIXME: whitelisting all cases is not great, try to improve
        for (origin, value) in auxiliary_imports:
            fout.write(f"from {origin} import {value}\n")

        with open(igraph.__file__, "rt") as f:
            # Swap in the method sources
            for line in f:
                mtype = None

                for mname in methodsources:
                    # Class methods (constructors)
                    if " " + mname + " = classmethod(_" in line:
                        mtype = "class"
                        break

                    # Instance methods (excluding factories e.g. 3d layouts)
                    if (" " + mname + " = _" in line) and ("(" not in line):
                        mtype = "instance"
                        break

                else:
                    fout.write(line)
                    continue

                # Method found, substitute and remove from dict
                fout.write("\n")
                if mtype == "class":
                    fout.write("    @classmethod\n")
                for mline in methodsources[mname]:
                    # Correct indentation
                    fout.write("    " + mline)

                del methodsources[mname]

    # Move the new file back
    with open(igraph.__file__, "wt") as fout:
        with open(newmodule, "rt") as f:
            fout.write(f.read())

    # Delete .new file
    os.remove(newmodule)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Update the contents of CONTRIBUTORS.txt based on .all-contributorsrc."""

from json import load
from os import chdir
from pathlib import Path


HEADER = """\
Thanks goes to these wonderful people:

"""

FOOTER = """\

This project follows the [all-contributors][1] specification. Contributions of any kind welcome!

This file is an automatically generated, plain-text version of CONTRIBUTORS.md.

[1]: https://github.com/all-contributors/all-contributors
"""

def main():
    root_dir = Path(__file__).parent.parent
    chdir(root_dir)

    with (root_dir / ".all-contributorsrc").open("r") as fp:
        contributors = load(fp)["contributors"]

    with (root_dir / "CONTRIBUTORS.txt").open("w") as fp:
        fp.write(HEADER)
        for c in contributors:
            if c["name"]:
                fp.write(f"{c['name']} (@{c['login']})\n")
            else:
                fp.write(f"@{c['login']}\n")
        fp.write(FOOTER)


if __name__ == "__main__":
    main()



"""
.. _tutorials-visualize-communities:

=====================
Communities
=====================

This example shows how to visualize communities or clusters of a graph.
"""
import igraph as ig
import matplotlib.pyplot as plt

# %%
# First, we generate a graph. We use a famous graph here for simplicity:
g = ig.Graph.Famous("Zachary")

# %%
# Edge betweenness is a standard way to detect communities. We then covert into
# a :class:`igraph.VertexClustering` object for subsequent ease of use:
communities = g.community_edge_betweenness()
communities = communities.as_clustering()

# %%
# Next, we color each vertex and edge based on its community membership:
num_communities = len(communities)
palette = ig.RainbowPalette(n=num_communities)
for i, community in enumerate(communities):
    g.vs[community]["color"] = i
    community_edges = g.es.select(_within=community)
    community_edges["color"] = i


# %%
# Last, we plot the graph. We use a fancy technique called proxy artists to
# make a legend. You can find more about that in matplotlib's
# :doc:`matplotlib:users/explain/axes/legend_guide`:
fig, ax = plt.subplots()
ig.plot(
    communities,
    palette=palette,
    edge_width=1,
    target=ax,
    vertex_size=20,
)

# Create a custom color legend
legend_handles = []
for i in range(num_communities):
    handle = ax.scatter(
        [], [],
        s=100,
        facecolor=palette.get(i),
        edgecolor="k",
        label=i,
    )
    legend_handles.append(handle)
ax.legend(
    handles=legend_handles,
    title='Community:',
    bbox_to_anchor=(0, 1.0),
    bbox_transform=ax.transAxes,
)
plt.show()

# %%
# For an example on how to generate the cluster graph from a vertex cluster,
# check out :ref:`tutorials-cluster-graph`.



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Additional auxiliary data types"""

__all__ = ("Matrix",)


class Matrix:
    """Simple matrix data type.

    Of course there are much more advanced matrix data types for Python (for
    instance, the C{ndarray} data type of Numeric Python) and this implementation
    does not want to compete with them. The only role of this data type is to
    provide a convenient interface for the matrices returned by the C{Graph}
    object (for instance, allow indexing with tuples in the case of adjacency
    matrices and so on).
    """

    def __init__(self, data=None):
        """Initializes a matrix.

        @param data: the elements of the matrix as a list of lists, or C{None} to
          create a 0x0 matrix.
        """
        self._nrow, self._ncol, self._data = 0, 0, []
        self.data = data

    @classmethod
    def Fill(cls, value, *args):
        """Creates a matrix filled with the given value

        @param value: the value to be used
        @keyword shape: the shape of the matrix. Can be a single integer,
          two integers or a tuple. If a single integer is
          given here, the matrix is assumed to be square-shaped.
        """
        if len(args) < 1:
            raise TypeError("expected an integer or a tuple")
        if len(args) == 1:
            if hasattr(args[0], "__len__"):
                height, width = int(args[0][0]), int(args[0][1])
            else:
                height, width = int(args[0]), int(args[0])
        else:
            height, width = int(args[0]), int(args[1])
        mtrx = [[value] * width for _ in range(height)]
        return cls(mtrx)

    @classmethod
    def Zero(cls, *args):
        """Creates a matrix filled with zeros.

        @keyword shape: the shape of the matrix. Can be a single integer,
          two integers or a tuple. If a single integer is
          given here, the matrix is assumed to be square-shaped.
        """
        result = cls.Fill(0, *args)
        return result

    @classmethod
    def Identity(cls, *args):
        """Creates an identity matrix.

        @keyword shape: the shape of the matrix. Can be a single integer,
          two integers or a tuple. If a single integer is
          given here, the matrix is assumed to be square-shaped.
        """
        result = cls.Fill(0, *args)
        for i in range(min(result.shape)):
            result._data[i][i] = 1
        return result

    def _set_data(self, data=None):
        """Sets the data stored in the matrix"""
        if data is not None:
            self._data = [list(row) for row in data]
            self._nrow = len(self._data)
            if self._nrow > 0:
                self._ncol = max(len(row) for row in self._data)
            else:
                self._ncol = 0
            for row in self._data:
                if len(row) < self._ncol:
                    row.extend([0] * (self._ncol - len(row)))

    def _get_data(self):
        """Returns the data stored in the matrix as a list of lists"""
        return [list(row) for row in self._data]

    data = property(_get_data, _set_data)

    @property
    def shape(self):
        """Returns the shape of the matrix as a tuple"""
        return self._nrow, self._ncol

    def __add__(self, other):
        """Adds the given value to the matrix.

        @param other: either a scalar or a matrix. Scalars will
          be added to each element of the matrix. Matrices will
          be added together elementwise.
        @return: the result matrix
        """
        if isinstance(other, Matrix):
            if self.shape != other.shape:
                raise ValueError("matrix shapes do not match")
            return self.__class__(
                [
                    [a + b for a, b in zip(row_a, row_b)]
                    for row_a, row_b in zip(self, other)
                ]
            )
        else:
            return self.__class__([[item + other for item in row] for row in self])

    def __eq__(self, other):
        """Checks whether a given matrix is equal to another one"""
        return (
            isinstance(other, Matrix)
            and self._nrow == other._nrow
            and self._ncol == other._ncol
            and self._data == other._data
        )

    def __getitem__(self, i):
        """Returns a single item, a row or a column of the matrix

        @param i: if a single integer, returns the M{i}th row as a list. If a
          slice, returns the corresponding rows as another L{Matrix} object. If
          a 2-tuple, the first element of the tuple is used to select a row and
          the second is used to select a column.
        """
        if isinstance(i, int):
            return list(self._data[i])
        elif isinstance(i, slice):
            return self.__class__(self._data[i])
        elif isinstance(i, tuple):
            try:
                first = i[0]
            except IndexError:
                first = slice(None)
            try:
                second = i[1]
            except IndexError:
                second = slice(None)
            if type(first) == slice and type(second) == slice:
                return self.__class__(row[second] for row in self._data[first])
            elif type(first) == slice:
                return [row[second] for row in self._data[first]]
            else:
                return self._data[first][second]
        else:
            raise IndexError("invalid matrix index")

    def __hash__(self):
        """Returns a hash value for a matrix."""
        return hash(self._nrow, self._ncol, self._data)

    def __iadd__(self, other):
        """In-place addition of a matrix or scalar."""
        if isinstance(other, Matrix):
            if self.shape != other.shape:
                raise ValueError("matrix shapes do not match")
            for row_a, row_b in zip(self._data, other):
                for i in range(len(row_a)):
                    row_a[i] += row_b[i]
        else:
            for row in self._data:
                for i in range(len(row)):
                    row[i] += other
        return self

    def __isub__(self, other):
        """In-place subtraction of a matrix or scalar."""
        if isinstance(other, Matrix):
            if self.shape != other.shape:
                raise ValueError("matrix shapes do not match")
            for row_a, row_b in zip(self._data, other):
                for i in range(len(row_a)):
                    row_a[i] -= row_b[i]
        else:
            for row in self._data:
                for i in range(len(row)):
                    row[i] -= other
        return self

    def __len__(self):
        """Returns the number of rows in the matrix."""
        return len(self._data)

    def __ne__(self, other):
        """Checks whether a given matrix is not equal to another one"""
        return not self == other

    def __setitem__(self, i, value):
        """Sets a single item, a row or a column of the matrix

        @param i: if a single integer, sets the M{i}th row as a list. If a
          slice, sets the corresponding rows from another L{Matrix} object.
          If a 2-tuple, the first element of the tuple is used to select a row
          and the second is used to select a column.
        @param value: the new value
        """
        if isinstance(i, int):
            # Setting a row
            if len(value) != len(self._data[i]):
                raise ValueError("new value must have %d items" % self._ncol)
            self._data[i] = list(value)
        elif isinstance(i, slice):
            # Setting multiple rows
            if len(value) != len(self._data[i]):
                raise ValueError("new value must have %d items" % self._ncol)
            if any(len(row) != self._ncol for row in value):
                raise ValueError("rows of new value must have %d items" % self._ncol)
            self._data[i] = [list(row) for row in value]
        elif isinstance(i, tuple):
            try:
                first = i[0]
            except IndexError:
                first = slice(None)
            try:
                second = i[1]
            except IndexError:
                second = slice(None)
            if type(first) == slice and type(second) == slice:
                # Setting a submatrix
                # TODO
                raise NotImplementedError
            elif type(first) == slice:
                # Setting a submatrix
                raise NotImplementedError
            else:
                # Setting a single element
                self._data[first][second] = value
        else:
            raise IndexError("invalid matrix index")

    def __sub__(self, other):
        """Subtracts the given value from the matrix.

        @param other: either a scalar or a matrix. Scalars will
          be subtracted from each element of the matrix. Matrices will
          be subtracted together elementwise.
        @return: the result matrix
        """
        if isinstance(other, Matrix):
            if self.shape != other.shape:
                raise ValueError("matrix shapes do not match")
            return self.__class__(
                [
                    [a - b for a, b in zip(row_a, row_b)]
                    for row_a, row_b in zip(self, other)
                ]
            )
        else:
            return self.__class__([[item - other for item in row] for row in self])

    def __repr__(self):
        class_name = self.__class__.__name__
        rows = ("[%s]" % ", ".join(repr(item) for item in row) for row in self)
        return "%s([%s])" % (class_name, ", ".join(rows))

    def __str__(self):
        rows = ("[%s]" % ", ".join(repr(item) for item in row) for row in self)
        return "[%s]" % "\n ".join(rows)

    def __iter__(self):
        """Support for iteration.

        This is actually implemented as a generator, so there is no need for a
        separate iterator class. The generator returns I{copies} of the rows in
        the matrix as lists to avoid messing around with the internals. Feel
        free to do anything with the copies, the changes won't be reflected in
        the original matrix."""
        return (list(row) for row in self._data)

    def __plot__(self, backend, context, **kwds):
        """Plots the matrix to the given Cairo context or matplotlib Axes.

        Besides the usual self-explanatory plotting parameters (C{context},
        C{bbox}, C{palette}), it accepts the following keyword arguments:

          - C{style}: the style of the plot. C{boolean} is useful for plotting
            matrices with boolean (C{True}/C{False} or 0/1) values: C{False}
            will be shown with a white box and C{True} with a black box.
            C{palette} uses the given palette to represent numbers by colors,
            the minimum will be assigned to palette color index 0 and the maximum
            will be assigned to the length of the palette. C{None} draws transparent
            cell backgrounds only. The default style is C{boolean} (but it may
            change in the future). C{None} values in the matrix are treated
            specially in both cases: nothing is drawn in the cell corresponding
            to C{None}.

          - C{square}: whether the cells of the matrix should be square or not.
            Default is C{True}.

          - C{grid_width}: line width of the grid shown on the matrix. If zero or
            negative, the grid is turned off. The grid is also turned off if the size
            of a cell is less than three times the given line width. Default is C{1}.
            Fractional widths are also allowed.

          - C{border_width}: line width of the border drawn around the matrix.
            If zero or negative, the border is turned off. Default is C{1}.

          - C{row_names}: the names of the rows

          - C{col_names}: the names of the columns.

          - C{values}: values to be displayed in the cells. If C{None} or
            C{False}, no values are displayed. If C{True}, the values come
            from the matrix being plotted. If it is another matrix, the
            values of that matrix are shown in the cells. In this case,
            the shape of the value matrix must match the shape of the
            matrix being plotted.

          - C{value_format}: a format string or a callable that specifies how
            the values should be plotted. If it is a callable, it must be a
            function that expects a single value and returns a string.
            Example: C{"%#.2f"} for floating-point numbers with always exactly
            two digits after the decimal point. See the Python documentation of
            the C{%} operator for details on the format string. If the format
            string is not given, it defaults to the C{str} function.

        If only the row names or the column names are given and the matrix
        is square-shaped, the same names are used for both column and row
        names.
        """
        from igraph.drawing import DrawerDirectory

        drawer = DrawerDirectory.resolve(self, backend)(context)
        drawer.draw(self, **kwds)

    def min(self, dim=None):
        """Returns the minimum of the matrix along the given dimension

        @param dim: the dimension. 0 means determining the column minimums, 1 means
          determining the row minimums. If C{None}, the global minimum is
          returned.
        """
        if dim == 1:
            return [min(row) for row in self._data]
        if dim == 0:
            return [min(row[idx] for row in self._data) for idx in range(self._ncol)]
        return min(min(row) for row in self._data)

    def max(self, dim=None):
        """Returns the maximum of the matrix along the given dimension

        @param dim: the dimension. 0 means determining the column maximums, 1 means
          determining the row maximums. If C{None}, the global maximum is
          returned.
        """
        if dim == 1:
            return [max(row) for row in self._data]
        if dim == 0:
            return [max(row[idx] for row in self._data) for idx in range(self._ncol)]
        return max(max(row) for row in self._data)


class DyadCensus(tuple):
    """Dyad census of a graph.

    This is a pretty simple class - basically it is a tuple, but it allows
    the user to refer to its individual items by the names C{mutual} (or
    C{mut}), C{asymmetric} (or C{asy} or C{asym} or C{asymm}) and C{null}.

    Examples:

      >>> from igraph import Graph
      >>> g=Graph.Erdos_Renyi(100, 0.2, directed=True)
      >>> dc=g.dyad_census()
      >>> print(dc.mutual)            #doctest:+SKIP
      179
      >>> print(dc["asym"])           #doctest:+SKIP
      1609
      >>> print(tuple(dc), list(dc))  #doctest:+SKIP
      (179, 1609, 3162) [179, 1609, 3162]
      >>> print(sorted(dc.as_dict().items()))  #doctest:+ELLIPSIS
      [('asymmetric', ...), ('mutual', ...), ('null', ...)]
    """

    _remap = {
        "mutual": 0,
        "mut": 0,
        "sym": 0,
        "symm": 0,
        "asy": 1,
        "asym": 1,
        "asymm": 1,
        "asymmetric": 1,
        "null": 2,
    }

    def __getitem__(self, idx):
        return tuple.__getitem__(self, self._remap.get(idx, idx))

    def __getattr__(self, attr):
        if attr in self._remap:
            return tuple.__getitem__(self, self._remap[attr])
        raise AttributeError("no such attribute: %s" % attr)

    def __repr__(self):
        return "DyadCensus((%d, %d, %d))" % self

    def __str__(self):
        return "%d mutual, %d asymmetric, %d null dyads" % self

    def as_dict(self):
        """Converts the dyad census to a dict using the known dyad names."""
        return {"mutual": self[0], "asymmetric": self[1], "null": self[2]}


class TriadCensus(tuple):
    """Triad census of a graph.

    This is a pretty simple class - basically it is a tuple, but it allows
    the user to refer to its individual items by the following triad names:

      - C{003} -- the empty graph
      - C{012} -- a graph with a single directed edge (C{A --> B, C})
      - C{102} -- a graph with a single mutual edge (C{A <-> B, C})
      - C{021D} -- the binary out-tree (C{A <-- B --> C})
      - C{021U} -- the binary in-tree (C{A --> B <-- C})
      - C{021C} -- the directed line (C{A --> B --> C})
      - C{111D} -- C{A <-> B <-- C}
      - C{111U} -- C{A <-> B --> C}
      - C{030T} -- C{A --> B <-- C, A --> C}
      - C{030C} -- C{A <-- B <-- C, A --> C}
      - C{201} -- C{A <-> B <-> C}
      - C{120D} -- C{A <-- B --> C, A <-> C}
      - C{120U} -- C{A --> B <-- C, A <-> C}
      - C{120C} -- C{A --> B --> C, A <-> C}
      - C{210C} -- C{A --> B <-> C, A <-> C}
      - C{300} -- the complete graph (C{A <-> B <-> C, A <-> C})

    Attribute and item accessors are provided. Due to the syntax of Python,
    attribute names are not allowed to start with a number, therefore the
    triad names must be prepended with a lowercase C{t} when accessing
    them as attributes. This is not necessary with the item accessor syntax.

    Examples:

      >>> from igraph import Graph
      >>> g=Graph.Erdos_Renyi(100, 0.2, directed=True)
      >>> tc=g.triad_census()
      >>> print(tc.t003)                    #doctest:+SKIP
      39864
      >>> print(tc["030C"])                 #doctest:+SKIP
      1206
    """

    _remap = {
        "003": 0,
        "012": 1,
        "102": 2,
        "021D": 3,
        "021U": 4,
        "021C": 5,
        "111D": 6,
        "111U": 7,
        "030T": 8,
        "030C": 9,
        "201": 10,
        "120D": 11,
        "120U": 12,
        "120C": 13,
        "210": 14,
        "300": 15,
    }

    def __getitem__(self, idx):
        if isinstance(idx, str):
            idx = idx.upper()
        return tuple.__getitem__(self, self._remap.get(idx, idx))

    def __getattr__(self, attr):
        if isinstance(attr, str) and attr[0] == "t" and attr[1:].upper() in self._remap:
            return tuple.__getitem__(self, self._remap[attr[1:].upper()])
        raise AttributeError("no such attribute: %s" % attr)

    def __repr__(self):
        return "TriadCensus((%s))" % ", ".join(str(item) for item in self)

    def __str__(self):
        maxidx = len(self)
        maxcount = max(self)
        numwidth = len(str(maxcount))
        captionwidth = max(len(key) for key in self._remap)
        colcount = 4

        rowcount = maxidx / colcount
        if rowcount * colcount < maxidx:
            rowcount += 1

        invmap = {v: k for k, v in self._remap.items()}
        result, row, idx = [], [], 0
        for _ in range(rowcount):
            for _ in range(colcount):
                if idx >= maxidx:
                    break
                row.append(
                    "%-*s: %*d"
                    % (captionwidth, invmap.get(idx, ""), numwidth, self[idx])
                )
                idx += 1
            result.append(" | ".join(row))
            row = []

        return "\n".join(result)


class UniqueIdGenerator:
    """A dictionary-like class that can be used to assign unique IDs to
    names (say, vertex names).

    Usage:

    >>> gen = UniqueIdGenerator()
    >>> gen["A"]
    0
    >>> gen["B"]
    1
    >>> gen["C"]
    2
    >>> gen["A"]      # Retrieving already existing ID
    0
    >>> gen.add("D")  # Synonym of gen["D"]
    3
    >>> len(gen)      # Number of already used IDs
    4
    >>> "C" in gen
    True
    >>> "E" in gen
    False
    """

    def __init__(self, id_generator=None, initial=None):
        """Creates a new unique ID generator. C{id_generator} specifies how do we
        assign new IDs to elements that do not have an ID yet. If it is C{None},
        elements will be assigned integer identifiers starting from 0. If it is
        an integer, elements will be assigned identifiers starting from the given
        integer. If it is an iterator or generator, its C{next()} method will be
        called every time a new ID is needed."""
        if id_generator is None:
            id_generator = 0
        if isinstance(id_generator, int):
            import itertools

            self._generator = itertools.count(id_generator)
        else:
            self._generator = id_generator
        self._ids = {}
        if initial:
            for value in initial:
                self.add(value)

    def __contains__(self, item):
        """Checks whether C{item} already has an ID or not."""
        return item in self._ids

    def __getitem__(self, item):
        """Retrieves the ID corresponding to C{item}. Generates a new ID for
        C{item} if it is the first time we request an ID for it."""
        try:
            return self._ids[item]
        except KeyError:
            self._ids[item] = next(self._generator)
            return self._ids[item]

    def __setitem__(self, item, value):
        """Overrides the ID for C{item}."""
        self._ids[item] = value

    def __len__(self):
        """Returns the number of items."""
        return len(self._ids)

    def reverse_dict(self):
        """Returns the reverse mapping, i.e., the one that maps from generated
        IDs to their corresponding objects"""
        return {v: k for k, v in self._ids.items()}

    def values(self):
        """Returns the values stored so far. If the generator generates items
        according to the standard sorting order, the values returned will be
        exactly in the order they were added. This holds for integer IDs for
        instance (but for many other ID generators as well)."""
        return sorted(self._ids.keys(), key=self._ids.__getitem__)

    add = __getitem__



# vim:set ts=4 sw=4 sts=4 et:
import unittest

from igraph import Graph


class SpectralTests(unittest.TestCase):
    def assertAlmostEqualMatrix(self, mat1, mat2, eps=1e-7):
        self.assertTrue(
            all(abs(obs - exp) < eps for obs, exp in zip(sum(mat1, []), sum(mat2, [])))
        )

    def testLaplacian(self):
        g = Graph.Full(3)
        g.es["weight"] = [1, 2, 3]
        self.assertTrue(g.laplacian() == [[2, -1, -1], [-1, 2, -1], [-1, -1, 2]])
        self.assertAlmostEqualMatrix(
            g.laplacian(normalized=True),
            [[1, -0.5, -0.5], [-0.5, 1, -0.5], [-0.5, -0.5, 1]],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(normalized="symmetric"),
            [[1, -0.5, -0.5], [-0.5, 1, -0.5], [-0.5, -0.5, 1]],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(normalized="left"),
            [[1, -0.5, -0.5], [-0.5, 1, -0.5], [-0.5, -0.5, 1]],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(normalized="right"),
            [[1, -0.5, -0.5], [-0.5, 1, -0.5], [-0.5, -0.5, 1]],
        )

        mx0 = [
            [1.0, -1 / (12**0.5), -2 / (15**0.5)],
            [-1 / (12**0.5), 1.0, -3 / (20**0.5)],
            [-2 / (15**0.5), -3 / (20**0.5), 1.0],
        ]
        self.assertAlmostEqualMatrix(g.laplacian("weight", True), mx0)

        g = Graph.Tree(5, 2)
        g.add_vertices(1)
        self.assertTrue(
            g.laplacian()
            == [
                [2, -1, -1, 0, 0, 0],
                [-1, 3, 0, -1, -1, 0],
                [-1, 0, 1, 0, 0, 0],
                [0, -1, 0, 1, 0, 0],
                [0, -1, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0],
            ]
        )

        g = Graph.Formula("A --> B --> C --> D --> E --> A, A --> C")
        self.assertAlmostEqualMatrix(
            g.laplacian(mode="out"),
            [
                [2, -1, -1, 0, 0],
                [0, 1, -1, 0, 0],
                [0, 0, 1, -1, 0],
                [0, 0, 0, 1, -1],
                [-1, 0, 0, 0, 1],
            ],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(mode="in"),
            [
                [1, -1, -1, 0, 0],
                [0, 1, -1, 0, 0],
                [0, 0, 2, -1, 0],
                [0, 0, 0, 1, -1],
                [-1, 0, 0, 0, 1],
            ],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(mode="out", normalized="left"),
            [
                [1, -0.5, -0.5, 0, 0],
                [0, 1, -1, 0, 0],
                [0, 0, 1, -1, 0],
                [0, 0, 0, 1, -1],
                [-1, 0, 0, 0, 1],
            ],
        )
        self.assertAlmostEqualMatrix(
            g.laplacian(mode="in", normalized="right"),
            [
                [1, -1, -0.5, 0, 0],
                [0, 1, -0.5, 0, 0],
                [0, 0, 1, -1, 0],
                [0, 0, 0, 1, -1],
                [-1, 0, 0, 0, 1],
            ],
        )


def suite():
    spectral_suite = unittest.defaultTestLoader.loadTestsFromTestCase(SpectralTests)
    return unittest.TestSuite([spectral_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



#!/usr/bin/env python3
import pyodide_build

from pathlib import Path
from urllib.request import urlretrieve

target_dir = Path(pyodide_build.__file__).parent / "tools" / "cmake" / "Modules" / "Platform"
target_dir.mkdir(exist_ok=True, parents=True)

target_file = target_dir / "Emscripten.cmake"
if not target_file.is_file():
    url = "https://raw.githubusercontent.com/pyodide/pyodide-build/main/pyodide_build/tools/cmake/Modules/Platform/Emscripten.cmake"
    urlretrieve(url, str(target_file))
